
from PySide2 import QtCore

qt_resource_data = b"\
\x00\x00\x18\x9c\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
<svg xmlns:dc=\x22h\
ttp://purl.org/d\
c/elements/1.1/\x22\
 xmlns:cc=\x22http:\
//creativecommon\
s.org/ns#\x22 xmlns\
:rdf=\x22http://www\
.w3.org/1999/02/\
22-rdf-syntax-ns\
#\x22 xmlns:svg=\x22ht\
tp://www.w3.org/\
2000/svg\x22 xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22 xml\
ns:xlink=\x22http:/\
/www.w3.org/1999\
/xlink\x22 xmlns:so\
dipodi=\x22http://s\
odipodi.sourcefo\
rge.net/DTD/sodi\
podi-0.dtd\x22 xmln\
s:inkscape=\x22http\
://www.inkscape.\
org/namespaces/i\
nkscape\x22 width=\x22\
64\x22 height=\x2264\x22 \
id=\x22svg2\x22 versio\
n=\x221.1\x22 inkscape\
:version=\x220.48.5\
 r10040\x22 sodipod\
i:docname=\x22previ\
ew-rendered.svg\x22\
>\x0a  <defs id=\x22de\
fs4\x22>\x0a    <linea\
rGradient id=\x22li\
nearGradient3755\
\x22>\x0a      <stop s\
tyle=\x22stop-color\
:#424242;stop-op\
acity:1;\x22 offset\
=\x220\x22 id=\x22stop375\
7\x22/>\x0a      <stop\
 style=\x22stop-col\
or:#9f9f9f;stop-\
opacity:1;\x22 offs\
et=\x221\x22 id=\x22stop3\
759\x22/>\x0a    </lin\
earGradient>\x0a   \
 <filter color-i\
nterpolation-fil\
ters=\x22sRGB\x22 inks\
cape:collect=\x22al\
ways\x22 id=\x22filter\
3871\x22 x=\x22-0.4536\
3569\x22 width=\x221.9\
072714\x22 y=\x22-0.45\
363569\x22 height=\x22\
1.9072714\x22>\x0a    \
  <feGaussianBlu\
r inkscape:colle\
ct=\x22always\x22 stdD\
eviation=\x229.8975\
056\x22 id=\x22feGauss\
ianBlur3873\x22/>\x0a \
   </filter>\x0a   \
 <linearGradient\
 inkscape:collec\
t=\x22always\x22 xlink\
:href=\x22#linearGr\
adient3600\x22 id=\x22\
linearGradient36\
06\x22 x1=\x2251.03728\
1\x22 y1=\x2247.692612\
\x22 x2=\x2214.872466\x22\
 y2=\x2210.397644\x22 \
gradientUnits=\x22u\
serSpaceOnUse\x22/>\
\x0a    <linearGrad\
ient id=\x22linearG\
radient3600\x22>\x0a  \
    <stop style=\
\x22stop-color:#555\
753;stop-opacity\
:1\x22 offset=\x220\x22 i\
d=\x22stop3602\x22/>\x0a \
     <stop style\
=\x22stop-color:#d3\
d7cf;stop-opacit\
y:1\x22 offset=\x221\x22 \
id=\x22stop3604\x22/>\x0a\
    </linearGrad\
ient>\x0a    <filte\
r inkscape:colle\
ct=\x22always\x22 id=\x22\
filter3805\x22 x=\x22-\
0.29648924\x22 widt\
h=\x221.5929785\x22 y=\
\x22-0.30309633\x22 he\
ight=\x221.6061927\x22\
>\x0a      <feGauss\
ianBlur inkscape\
:collect=\x22always\
\x22 stdDeviation=\x22\
2.2490136\x22 id=\x22f\
eGaussianBlur380\
7\x22/>\x0a    </filte\
r>\x0a  </defs>\x0a  <\
sodipodi:namedvi\
ew id=\x22base\x22 pag\
ecolor=\x22#ffffff\x22\
 bordercolor=\x22#6\
66666\x22 borderopa\
city=\x221.0\x22 inksc\
ape:pageopacity=\
\x220.0\x22 inkscape:p\
ageshadow=\x222\x22 in\
kscape:zoom=\x224.8\
193618\x22 inkscape\
:cx=\x22-4.0027486\x22\
 inkscape:cy=\x2228\
.204234\x22 inkscap\
e:document-units\
=\x22px\x22 inkscape:c\
urrent-layer=\x22la\
yer1\x22 showgrid=\x22\
true\x22 showborder\
=\x22true\x22 inkscape\
:window-width=\x228\
00\x22 inkscape:win\
dow-height=\x22837\x22\
 inkscape:window\
-x=\x220\x22 inkscape:\
window-y=\x2227\x22 in\
kscape:window-ma\
ximized=\x220\x22>\x0a   \
 <inkscape:grid \
type=\x22xygrid\x22 id\
=\x22grid2988\x22 emps\
pacing=\x222\x22 visib\
le=\x22true\x22 enable\
d=\x22true\x22 snapvis\
iblegridlinesonl\
y=\x22true\x22/>\x0a  </s\
odipodi:namedvie\
w>\x0a  <metadata i\
d=\x22metadata7\x22>\x0a \
   <rdf:RDF>\x0a   \
   <cc:Work rdf:\
about=\x22\x22>\x0a      \
  <dc:format>ima\
ge/svg+xml</dc:f\
ormat>\x0a        <\
dc:type rdf:reso\
urce=\x22http://pur\
l.org/dc/dcmityp\
e/StillImage\x22/>\x0a\
        <dc:titl\
e/>\x0a        <dc:\
creator>\x0a       \
   <cc:Agent>\x0a  \
          <dc:ti\
tle>$committer</\
dc:title>\x0a      \
    </cc:Agent>\x0a\
        </dc:cre\
ator>\x0a        <d\
c:title>preview-\
rendered</dc:tit\
le>\x0a        <dc:\
date>2013-11-19<\
/dc:date>\x0a      \
  <dc:relation>h\
ttp://www.freeca\
d.org/wiki/index\
.php?title=Artwo\
rk</dc:relation>\
\x0a        <dc:pub\
lisher>\x0a        \
  <cc:Agent>\x0a   \
         <dc:tit\
le>FreeCAD</dc:t\
itle>\x0a          \
</cc:Agent>\x0a    \
    </dc:publish\
er>\x0a        <dc:\
identifier>FreeC\
AD/src/Mod/Mater\
ial/Resources/ic\
ons/preview-rend\
ered.svg</dc:ide\
ntifier>\x0a       \
 <dc:rights>\x0a   \
       <cc:Agent\
>\x0a            <d\
c:title>FreeCAD \
LGPL2+</dc:title\
>\x0a          </cc\
:Agent>\x0a        \
</dc:rights>\x0a   \
     <cc:license\
>https://www.gnu\
.org/copyleft/le\
sser.html</cc:li\
cense>\x0a        <\
dc:contributor>\x0a\
          <cc:Ag\
ent>\x0a           \
 <dc:title>[agry\
son] Alexander G\
ryson</dc:title>\
\x0a          </cc:\
Agent>\x0a        <\
/dc:contributor>\
\x0a      </cc:Work\
>\x0a    </rdf:RDF>\
\x0a  </metadata>\x0a \
 <g inkscape:lab\
el=\x22Layer 1\x22 ink\
scape:groupmode=\
\x22layer\x22 id=\x22laye\
r1\x22 transform=\x22t\
ranslate(0,-988.\
36218)\x22>\x0a    <pa\
th sodipodi:type\
=\x22arc\x22 style=\x22co\
lor:#000000;fill\
:#000000;fill-op\
acity:0.65625;fi\
ll-rule:evenodd;\
stroke:none;stro\
ke-width:5;marke\
r:none;visibilit\
y:visible;displa\
y:inline;overflo\
w:visible;filter\
:url(#filter3871\
);enable-backgro\
und:accumulate\x22 \
id=\x22path2826-8\x22 \
sodipodi:cx=\x2231.\
636364\x22 sodipodi\
:cy=\x2230.363636\x22 \
sodipodi:rx=\x2226.\
181818\x22 sodipodi\
:ry=\x2226.181818\x22 \
d=\x22m 57.818182,3\
0.363636 a 26.18\
1818,26.181818 0\
 1 1 -52.363636,\
0 26.181818,26.1\
81818 0 1 1 52.3\
63636,0 z\x22 trans\
form=\x22matrix(0.4\
8061681,0,0,0.12\
01542,24.795032,\
1042.7139)\x22/>\x0a  \
  <path sodipodi\
:type=\x22arc\x22 styl\
e=\x22color:#000000\
;fill:url(#linea\
rGradient3606);f\
ill-opacity:1;fi\
ll-rule:evenodd;\
stroke:#0b1521;s\
troke-width:1.80\
564272;marker:no\
ne;visibility:vi\
sible;display:in\
line;overflow:vi\
sible;enable-bac\
kground:accumula\
te\x22 id=\x22path2826\
\x22 sodipodi:cx=\x223\
1.636364\x22 sodipo\
di:cy=\x2230.363636\
\x22 sodipodi:rx=\x222\
6.181818\x22 sodipo\
di:ry=\x2226.181818\
\x22 d=\x22m 57.818182\
,30.363636 a 26.\
181818,26.181818\
 0 1 1 -52.36363\
6,0 26.181818,26\
.181818 0 1 1 52\
.363636,0 z\x22 tra\
nsform=\x22matrix(1\
.1076389,0,0,1.1\
076389,-3.041667\
2,986.73024)\x22/>\x0a\
    <path sodipo\
di:type=\x22arc\x22 st\
yle=\x22color:#0000\
00;fill:none;str\
oke:#000000;stro\
ke-width:1.93939\
412000000000;mar\
ker:none;visibil\
ity:visible;disp\
lay:inline;overf\
low:visible;enab\
le-background:ac\
cumulate\x22 id=\x22pa\
th2826-4\x22 sodipo\
di:cx=\x2231.636364\
\x22 sodipodi:cy=\x223\
0.363636\x22 sodipo\
di:rx=\x2226.181818\
\x22 sodipodi:ry=\x222\
6.181818\x22 d=\x22m 5\
7.818182,30.3636\
36 a 26.181818,2\
6.181818 0 1 1 -\
52.363636,0 26.1\
81818,26.181818 \
0 1 1 52.363636,\
0 z\x22 transform=\x22\
matrix(1.03125,0\
,0,1.03125,-0.62\
500046,989.04968\
)\x22/>\x0a    <path s\
odipodi:type=\x22ar\
c\x22 style=\x22opacit\
y:0.844827569999\
99986;color:#000\
000;fill:#fffffe\
;fill-opacity:1;\
fill-rule:evenod\
d;stroke:none;st\
roke-width:5;mar\
ker:none;visibil\
ity:visible;disp\
lay:inline;overf\
low:visible;enab\
le-background:ac\
cumulate;filter:\
url(#filter3805)\
\x22 id=\x22path3608\x22 \
sodipodi:cx=\x2211.\
181818\x22 sodipodi\
:cy=\x2220.181818\x22 \
sodipodi:rx=\x2210.\
454545\x22 sodipodi\
:ry=\x227.2727275\x22 \
d=\x22m 21.636363,2\
0.181818 a 10.45\
4545,7.2727275 0\
 1 1 -20.9090900\
1,0 10.454545,7.\
2727275 0 1 1 20\
.90909001,0 z\x22 t\
ransform=\x22matrix\
(0.85867478,-0.7\
467709,0.7057652\
6,0.90856459,-4.\
8393203,997.4188\
)\x22/>\x0a    <path s\
odipodi:type=\x22ar\
c\x22 style=\x22color:\
#000000;fill:non\
e;stroke:#2e3436\
;stroke-width:1.\
8056427200000000\
3;marker:none;vi\
sibility:visible\
;display:inline;\
overflow:visible\
;enable-backgrou\
nd:accumulate\x22 i\
d=\x22path2826-5\x22 s\
odipodi:cx=\x2231.6\
36364\x22 sodipodi:\
cy=\x2230.363636\x22 s\
odipodi:rx=\x2226.1\
81818\x22 sodipodi:\
ry=\x2226.181818\x22 d\
=\x22m 57.818182,30\
.363636 a 26.181\
818,26.181818 0 \
1 1 -52.363636,0\
 26.181818,26.18\
1818 0 1 1 52.36\
3636,0 z\x22 transf\
orm=\x22matrix(1.10\
76389,0,0,1.1076\
389,-3.0416672,9\
86.73024)\x22/>\x0a  <\
/g>\x0a</svg>\x0a\
\x00\x00\x16\xbd\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:xlink=\x22htt\
p://www.w3.org/1\
999/xlink\x22\x0a   xm\
lns:sodipodi=\x22ht\
tp://sodipodi.so\
urceforge.net/DT\
D/sodipodi-0.dtd\
\x22\x0a   xmlns:inksc\
ape=\x22http://www.\
inkscape.org/nam\
espaces/inkscape\
\x22\x0a   width=\x2264\x22\x0a\
   height=\x2264\x22\x0a \
  id=\x22svg2\x22\x0a   v\
ersion=\x221.1\x22\x0a   \
inkscape:version\
=\x220.48.5 r10040\x22\
\x0a   sodipodi:doc\
name=\x22preview-ve\
ctor.svg\x22>\x0a  <de\
fs\x0a     id=\x22defs\
4\x22>\x0a    <inkscap\
e:perspective\x0a  \
     sodipodi:ty\
pe=\x22inkscape:per\
sp3d\x22\x0a       ink\
scape:vp_x=\x22-206\
.25075 : 119.688\
54 : 1\x22\x0a       i\
nkscape:vp_y=\x220 \
: 1000 : 0\x22\x0a    \
   inkscape:vp_z\
=\x22183.50001 : 14\
0.32143 : 1\x22\x0a   \
    inkscape:per\
sp3d-origin=\x2248 \
: 32 : 1\x22\x0a      \
 id=\x22perspective\
3792\x22 />\x0a    <li\
nearGradient\x0a   \
    id=\x22linearGr\
adient3755\x22>\x0a   \
   <stop\x0a       \
  style=\x22stop-co\
lor:#424242;stop\
-opacity:1;\x22\x0a   \
      offset=\x220\x22\
\x0a         id=\x22st\
op3757\x22 />\x0a     \
 <stop\x0a         \
style=\x22stop-colo\
r:#9f9f9f;stop-o\
pacity:1;\x22\x0a     \
    offset=\x221\x22\x0a \
        id=\x22stop\
3759\x22 />\x0a    </l\
inearGradient>\x0a \
   <linearGradie\
nt\x0a       inksca\
pe:collect=\x22alwa\
ys\x22\x0a       xlink\
:href=\x22#linearGr\
adient3755\x22\x0a    \
   id=\x22linearGra\
dient3761\x22\x0a     \
  x1=\x2275.724594\x22\
\x0a       y1=\x2285.0\
23956\x22\x0a       x2\
=\x2241.895164\x22\x0a   \
    y2=\x2213.33485\
1\x22\x0a       gradie\
ntUnits=\x22userSpa\
ceOnUse\x22 />\x0a  </\
defs>\x0a  <sodipod\
i:namedview\x0a    \
 id=\x22base\x22\x0a     \
pagecolor=\x22#ffff\
ff\x22\x0a     borderc\
olor=\x22#666666\x22\x0a \
    borderopacit\
y=\x221.0\x22\x0a     ink\
scape:pageopacit\
y=\x220.0\x22\x0a     ink\
scape:pageshadow\
=\x222\x22\x0a     inksca\
pe:zoom=\x228.75711\
68\x22\x0a     inkscap\
e:cx=\x2239.859065\x22\
\x0a     inkscape:c\
y=\x2230.925664\x22\x0a  \
   inkscape:docu\
ment-units=\x22px\x22\x0a\
     inkscape:cu\
rrent-layer=\x22g37\
94\x22\x0a     showgri\
d=\x22true\x22\x0a     sh\
owborder=\x22true\x22\x0a\
     inkscape:wi\
ndow-width=\x22800\x22\
\x0a     inkscape:w\
indow-height=\x2283\
6\x22\x0a     inkscape\
:window-x=\x220\x22\x0a  \
   inkscape:wind\
ow-y=\x2227\x22\x0a     i\
nkscape:window-m\
aximized=\x220\x22\x0a   \
  inkscape:snap-\
bbox=\x22false\x22\x0a   \
  inkscape:snap-\
nodes=\x22true\x22\x0a   \
  inkscape:snap-\
global=\x22true\x22>\x0a \
   <inkscape:gri\
d\x0a       type=\x22x\
ygrid\x22\x0a       id\
=\x22grid3002\x22\x0a    \
   empspacing=\x222\
\x22\x0a       visible\
=\x22true\x22\x0a       e\
nabled=\x22true\x22\x0a  \
     snapvisible\
gridlinesonly=\x22t\
rue\x22 />\x0a  </sodi\
podi:namedview>\x0a\
  <metadata\x0a    \
 id=\x22metadata7\x22>\
\x0a    <rdf:RDF>\x0a \
     <cc:Work\x0a  \
       rdf:about\
=\x22\x22>\x0a        <dc\
:format>image/sv\
g+xml</dc:format\
>\x0a        <dc:ty\
pe\x0a           rd\
f:resource=\x22http\
://purl.org/dc/d\
cmitype/StillIma\
ge\x22 />\x0a        <\
dc:title />\x0a    \
    <dc:creator>\
\x0a          <cc:A\
gent>\x0a          \
  <dc:title>$com\
mitter</dc:title\
>\x0a          </cc\
:Agent>\x0a        \
</dc:creator>\x0a  \
      <dc:title>\
preview-vector</\
dc:title>\x0a      \
  <dc:date>2013-\
11-19</dc:date>\x0a\
        <dc:rela\
tion>http://www.\
freecad.org/wiki\
/index.php?title\
=Artwork</dc:rel\
ation>\x0a        <\
dc:publisher>\x0a  \
        <cc:Agen\
t>\x0a            <\
dc:title>FreeCAD\
</dc:title>\x0a    \
      </cc:Agent\
>\x0a        </dc:p\
ublisher>\x0a      \
  <dc:identifier\
>FreeCAD/src/Mod\
/Material/Resour\
ces/icons/previe\
w-vector.svg</dc\
:identifier>\x0a   \
     <dc:rights>\
\x0a          <cc:A\
gent>\x0a          \
  <dc:title>Free\
CAD LGPL2+</dc:t\
itle>\x0a          \
</cc:Agent>\x0a    \
    </dc:rights>\
\x0a        <cc:lic\
ense>https://www\
.gnu.org/copylef\
t/lesser.html</c\
c:license>\x0a     \
   <dc:contribut\
or>\x0a          <c\
c:Agent>\x0a       \
     <dc:title>[\
agryson] Alexand\
er Gryson</dc:ti\
tle>\x0a          <\
/cc:Agent>\x0a     \
   </dc:contribu\
tor>\x0a      </cc:\
Work>\x0a    </rdf:\
RDF>\x0a  </metadat\
a>\x0a  <g\x0a     ink\
scape:label=\x22Lay\
er 1\x22\x0a     inksc\
ape:groupmode=\x22l\
ayer\x22\x0a     id=\x22l\
ayer1\x22\x0a     tran\
sform=\x22translate\
(0,-988.36218)\x22>\
\x0a    <g\x0a       i\
d=\x22g3794\x22\x0a      \
 style=\x22color:#0\
00000;fill:#ffff\
ff;fill-opacity:\
1;fill-rule:even\
odd;stroke:none;\
stroke-width:3.2\
9399991;marker:n\
one;visibility:v\
isible;display:i\
nline;overflow:v\
isible;enable-ba\
ckground:accumul\
ate\x22>\x0a      <pat\
h\x0a         style\
=\x22fill:#d3d7cf;f\
ill-opacity:1;fi\
ll-rule:evenodd;\
stroke:#000000;s\
troke-width:2;st\
roke-linecap:rou\
nd;stroke-linejo\
in:round;stroke-\
miterlimit:4;str\
oke-opacity:1;st\
roke-dasharray:n\
one\x22\x0a         d=\
\x22m 29.814326,991\
.36218 -8.814326\
,4 -1,7.00002 26\
,1 14.994091,-8.\
6577 z\x22\x0a        \
 id=\x22path3816\x22\x0a \
        inkscape\
:connector-curva\
ture=\x220\x22\x0a       \
  sodipodi:nodet\
ypes=\x22cccccc\x22 />\
\x0a      <path\x0a   \
      d=\x22M 38.48\
4871,1007.8171 4\
0,1046.3622 l 21\
,-22.4044 0,-29.\
25647 z\x22\x0a       \
  style=\x22fill:#b\
7b7b7;fill-opaci\
ty:1;fill-rule:e\
venodd;stroke:#0\
00000;stroke-wid\
th:2;stroke-line\
cap:round;stroke\
-linejoin:round;\
stroke-miterlimi\
t:4;stroke-opaci\
ty:1;stroke-dash\
array:none\x22\x0a    \
     id=\x22path381\
8\x22\x0a         inks\
cape:connector-c\
urvature=\x220\x22\x0a   \
      sodipodi:n\
odetypes=\x22ccccc\x22\
 />\x0a      <path\x0a\
         style=\x22\
fill:#555753;str\
oke:#555753;stro\
ke-width:2;strok\
e-linecap:butt;s\
troke-linejoin:m\
iter;stroke-opac\
ity:1\x22\x0a         \
d=\x22M 59.035745,9\
.6627807 59.1080\
06,34.818961 40.\
91531,54.264958 \
42.698526,19.372\
964 z\x22\x0a         \
id=\x22path3022\x22\x0a  \
       inkscape:\
connector-curvat\
ure=\x220\x22\x0a        \
 transform=\x22tran\
slate(0,988.3621\
8)\x22\x0a         sod\
ipodi:nodetypes=\
\x22ccccc\x22 />\x0a     \
 <path\x0a         \
d=\x22m 21,995.3621\
8 31,5.00002 -12\
,46 -30,-8 z\x22\x0a  \
       style=\x22fi\
ll:#ffffff;fill-\
opacity:1;fill-r\
ule:evenodd;stro\
ke:#000000;strok\
e-width:4;stroke\
-linejoin:round;\
stroke-miterlimi\
t:4;stroke-opaci\
ty:1;stroke-dash\
array:none\x22\x0a    \
     id=\x22path382\
0\x22\x0a         inks\
cape:connector-c\
urvature=\x220\x22\x0a   \
      sodipodi:n\
odetypes=\x22ccccc\x22\
 />\x0a      <path\x0a\
         d=\x22m 21\
,995.36218 31,5.\
00002 -12,46 -30\
,-8 z\x22\x0a         \
style=\x22color:#00\
0000;fill:none;s\
troke:#888a85;st\
roke-width:2;str\
oke-linejoin:rou\
nd;stroke-miterl\
imit:4;stroke-op\
acity:1;stroke-d\
asharray:none;ma\
rker:none;visibi\
lity:visible;dis\
play:inline;over\
flow:visible;ena\
ble-background:a\
ccumulate\x22\x0a     \
    id=\x22path3820\
-3\x22\x0a         ink\
scape:connector-\
curvature=\x220\x22\x0a  \
       sodipodi:\
nodetypes=\x22ccccc\
\x22 />\x0a    </g>\x0a  \
</g>\x0a</svg>\x0a\
"

qt_resource_name = b"\
\x00\x05\
\x00o\xa6S\
\x00i\
\x00c\x00o\x00n\x00s\
\x00\x14\
\x02\x9c\x05\x07\
\x00p\
\x00r\x00e\x00v\x00i\x00e\x00w\x00-\x00r\x00e\x00n\x00d\x00e\x00r\x00e\x00d\x00.\
\x00s\x00v\x00g\
\x00\x12\
\x0eO\xfeG\
\x00p\
\x00r\x00e\x00v\x00i\x00e\x00w\x00-\x00v\x00e\x00c\x00t\x00o\x00r\x00.\x00s\x00v\
\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x01\x8d\xe3s\x22p\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x02\
\x00\x00\x01\x8d\xe3s\x22p\
\x00\x00\x00\x10\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01\x8d\xe3s\x22p\
\x00\x00\x00>\x00\x00\x00\x00\x00\x01\x00\x00\x18\xa0\
\x00\x00\x01\x8d\xe3s\x22p\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Class to manage installation of sets of Python dependencies."""

import os
import subprocess
from typing import List

import addonmanager_freecad_interface as fci
from addonmanager_pyside_interface import QObject, Signal, is_interruption_requested

import addonmanager_utilities as utils
from addonmanager_installer import AddonInstaller, MacroInstaller
from Addon import Addon

translate = fci.translate


class DependencyInstaller(QObject):
    """Install Python dependencies using pip. Intended to be instantiated and then moved into a
    QThread: connect the run() function to the QThread's started() signal."""

    no_python_exe = Signal()
    no_pip = Signal(str)  # Attempted command
    failure = Signal(str, str)  # Short message, detailed message
    finished = Signal()

    def __init__(
        self,
        addons: List[Addon],
        python_requires: List[str],
        python_optional: List[str],
        location: os.PathLike = None,
    ):
        """Install the various types of dependencies that might be specified. If an optional
         dependency fails this is non-fatal, but other failures are considered fatal. If location
        is specified it overrides the FreeCAD user base directory setting: this is used mostly
        for testing purposes and shouldn't be set by normal code in most circumstances.
        """
        super().__init__()
        self.addons = addons
        self.python_requires = python_requires
        self.python_optional = python_optional
        self.location = location

    def run(self):
        """Normally not called directly, but rather connected to the worker thread's started
        signal."""
        if self._verify_pip():
            if self.python_requires or self.python_optional:
                if not is_interruption_requested():
                    self._install_python_packages()
        if not is_interruption_requested():
            self._install_addons()
        self.finished.emit()

    def _install_python_packages(self):
        """Install required and optional Python dependencies using pip."""

        if self.location:
            vendor_path = os.path.join(self.location, "AdditionalPythonPackages")
        else:
            vendor_path = utils.get_pip_target_directory()
        if not os.path.exists(vendor_path):
            os.makedirs(vendor_path)

        self._install_required(vendor_path)
        self._install_optional(vendor_path)

    def _verify_pip(self) -> bool:
        """Ensure that pip is working -- returns True if it is, or False if not. Also emits the
        no_pip signal if pip cannot execute."""
        python_exe = self._get_python()
        if not python_exe:
            return False
        try:
            proc = self._run_pip(["--version"])
            fci.Console.PrintMessage(proc.stdout + "\n")
        except subprocess.CalledProcessError:
            self.no_pip.emit(f"{python_exe} -m pip --version")
            return False
        return True

    def _install_required(self, vendor_path: str) -> bool:
        """Install the required Python package dependencies. If any fail a failure
        signal is emitted and the function exits without proceeding with any additional
        installations."""
        for pymod in self.python_requires:
            if is_interruption_requested():
                return False
            try:
                proc = self._run_pip(
                    [
                        "install",
                        "--disable-pip-version-check",
                        "--target",
                        vendor_path,
                        pymod,
                    ]
                )
                fci.Console.PrintMessage(proc.stdout + "\n")
            except subprocess.CalledProcessError as e:
                fci.Console.PrintError(str(e) + "\n")
                self.failure.emit(
                    translate(
                        "AddonsInstaller",
                        "Installation of Python package {} failed",
                    ).format(pymod),
                    str(e),
                )
                return False
        return True

    def _install_optional(self, vendor_path: str):
        """Install the optional Python package dependencies. If any fail a message is printed to
        the console, but installation of the others continues."""
        for pymod in self.python_optional:
            if is_interruption_requested():
                return
            try:
                proc = self._run_pip(
                    [
                        "install",
                        "--disable-pip-version-check",
                        "--target",
                        vendor_path,
                        pymod,
                    ]
                )
                fci.Console.PrintMessage(proc.stdout + "\n")
            except subprocess.CalledProcessError as e:
                fci.Console.PrintError(
                    translate(
                        "AddonsInstaller", "Installation of optional package failed"
                    )
                    + ":\n"
                    + str(e)
                    + "\n"
                )

    def _run_pip(self, args):
        python_exe = self._get_python()
        final_args = [python_exe, "-m", "pip"]
        final_args.extend(args)
        return self._subprocess_wrapper(final_args)

    @staticmethod
    def _subprocess_wrapper(args) -> subprocess.CompletedProcess:
        """Wrap subprocess call so test code can mock it."""
        return utils.run_interruptable_subprocess(args)

    def _get_python(self) -> str:
        """Wrap Python access so test code can mock it."""
        python_exe = utils.get_python_exe()
        if not python_exe:
            self.no_python_exe.emit()
        return python_exe

    def _install_addons(self):
        for addon in self.addons:
            if is_interruption_requested():
                return
            fci.Console.PrintMessage(
                translate(
                    "AddonsInstaller", "Installing required dependency {}"
                ).format(addon.name)
                + "\n"
            )
            if addon.macro is None:
                installer = AddonInstaller(addon)
            else:
                installer = MacroInstaller(addon)
            result = (
                installer.run()
            )  # Run in this thread, which should be off the GUI thread
            if not result:
                self.failure.emit(
                    translate(
                        "AddonsInstaller", "Installation of Addon {} failed"
                    ).format(addon.name),
                    "",
                )
                return



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Class to display a first-run dialog for the Addon Manager """

import os

from PySide import QtCore, QtWidgets

import FreeCAD
import FreeCADGui

import addonmanager_utilities as utils

# pylint: disable=too-few-public-methods


class FirstRunDialog:
    """Manage the display of the Addon Manager's first-run dialog, setting up some user
    preferences and making sure they are aware that this connects to the internet, downloads
    data, and possibly installs things that run code not affiliated with FreeCAD itself."""

    def __init__(self):
        self.pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Addons")
        self.readWarning = self.pref.GetBool("readWarning2022", False)

    def exec(self) -> bool:
        """Display a first-run dialog if needed, and return True to indicate the Addon Manager
        should continue loading, or False if the user cancelled the dialog and wants to exit."""
        if not self.readWarning:
            warning_dialog = FreeCADGui.PySideUic.loadUi(
                os.path.join(os.path.dirname(__file__), "first_run.ui")
            )
            warning_dialog.setWindowFlag(QtCore.Qt.WindowStaysOnTopHint, True)
            autocheck = self.pref.GetBool("AutoCheck", False)
            download_macros = self.pref.GetBool("DownloadMacros", False)
            proxy_string = self.pref.GetString("ProxyUrl", "")
            if self.pref.GetBool("NoProxyCheck", True):
                proxy_option = 0
            elif self.pref.GetBool("SystemProxyCheck", False):
                proxy_option = 1
            elif self.pref.GetBool("UserProxyCheck", False):
                proxy_option = 2

            def toggle_proxy_list(option: int):
                if option == 2:
                    warning_dialog.lineEditProxy.show()
                else:
                    warning_dialog.lineEditProxy.hide()

            warning_dialog.checkBoxAutoCheck.setChecked(autocheck)
            warning_dialog.checkBoxDownloadMacroMetadata.setChecked(download_macros)
            warning_dialog.comboBoxProxy.setCurrentIndex(proxy_option)
            toggle_proxy_list(proxy_option)
            if proxy_option == 2:
                warning_dialog.lineEditProxy.setText(proxy_string)

            warning_dialog.comboBoxProxy.currentIndexChanged.connect(toggle_proxy_list)

            warning_dialog.labelWarning.setStyleSheet(
                f"color:{utils.warning_color_string()};font-weight:bold;"
            )

            if warning_dialog.exec() == QtWidgets.QDialog.Accepted:
                self.readWarning = True
                self.pref.SetBool("readWarning2022", True)
                self.pref.SetBool(
                    "AutoCheck", warning_dialog.checkBoxAutoCheck.isChecked()
                )
                self.pref.SetBool(
                    "DownloadMacros",
                    warning_dialog.checkBoxDownloadMacroMetadata.isChecked(),
                )
                selected_proxy_option = warning_dialog.comboBoxProxy.currentIndex()
                if selected_proxy_option == 0:
                    self.pref.SetBool("NoProxyCheck", True)
                    self.pref.SetBool("SystemProxyCheck", False)
                    self.pref.SetBool("UserProxyCheck", False)
                elif selected_proxy_option == 1:
                    self.pref.SetBool("NoProxyCheck", False)
                    self.pref.SetBool("SystemProxyCheck", True)
                    self.pref.SetBool("UserProxyCheck", False)
                else:
                    self.pref.SetBool("NoProxyCheck", False)
                    self.pref.SetBool("SystemProxyCheck", False)
                    self.pref.SetBool("UserProxyCheck", True)
                    self.pref.SetString("ProxyUrl", warning_dialog.lineEditProxy.text())
        return self.readWarning



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022-2023 FreeCAD Project Association                   *
# *   Copyright (c) 2018 Gaël Écorchard <galou_breizh@yahoo.fr>             *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Utilities to work across different platforms, providers and python versions """

import os
import platform
import shutil
import stat
import subprocess
import re
import ctypes
from typing import Optional, Any

from urllib.parse import urlparse

try:
    from PySide import QtCore, QtWidgets
except ImportError:
    QtCore = None
    QtWidgets = None

import addonmanager_freecad_interface as fci

if fci.FreeCADGui:

    # If the GUI is up, we can use the NetworkManager to handle our downloads. If there is no event
    # loop running this is not possible, so fall back to requests (if available), or the native
    # Python urllib.request (if requests is not available).
    import NetworkManager  # Requires an event loop, so is only available with the GUI
else:
    try:
        import requests
    except ImportError:
        requests = None
        import urllib.request
        import ssl

#  @package AddonManager_utilities
#  \ingroup ADDONMANAGER
#  \brief Utilities to work across different platforms, providers and python versions
#  @{


translate = fci.translate


class ProcessInterrupted(RuntimeError):
    """An interruption request was received and the process killed because of it."""


def symlink(source, link_name):
    """Creates a symlink of a file, if possible. Note that it fails on most modern Windows
    installations"""

    if os.path.exists(link_name) or os.path.lexists(link_name):
        pass
    else:
        os_symlink = getattr(os, "symlink", None)
        if callable(os_symlink):
            os_symlink(source, link_name)
        else:
            # NOTE: This does not work on most normal Windows 10 and later installations, unless
            # developer mode is turned on. Make sure to catch any exception thrown and have a
            # fallback plan.
            csl = ctypes.windll.kernel32.CreateSymbolicLinkW
            csl.argtypes = (ctypes.c_wchar_p, ctypes.c_wchar_p, ctypes.c_uint32)
            csl.restype = ctypes.c_ubyte
            flags = 1 if os.path.isdir(source) else 0
            # set the SYMBOLIC_LINK_FLAG_ALLOW_UNPRIVILEGED_CREATE flag
            # (see https://blogs.windows.com/buildingapps/2016/12/02/symlinks-windows-10)
            flags += 2
            if csl(link_name, source, flags) == 0:
                raise ctypes.WinError()


def rmdir(path: os.PathLike) -> bool:
    try:
        if os.path.islink(path):
            os.unlink(path)  # Remove symlink
        else:
            shutil.rmtree(path, onerror=remove_readonly)
    except (WindowsError, PermissionError, OSError):
        return False
    return True


def remove_readonly(func, path, _) -> None:
    """Remove a read-only file."""

    os.chmod(path, stat.S_IWRITE)
    func(path)


def update_macro_details(old_macro, new_macro):
    """Update a macro with information from another one

    Update a macro with information from another one, supposedly the same but
    from a different source. The first source is supposed to be git, the second
    one the wiki.
    """

    if old_macro.on_git and new_macro.on_git:
        fci.Console.PrintLog(
            f'The macro "{old_macro.name}" is present twice in github, please report'
        )
    # We don't report macros present twice on the wiki because a link to a
    # macro is considered as a macro. For example, 'Perpendicular To Wire'
    # appears twice, as of 2018-05-05).
    old_macro.on_wiki = new_macro.on_wiki
    for attr in ["desc", "url", "code"]:
        if not hasattr(old_macro, attr):
            setattr(old_macro, attr, getattr(new_macro, attr))


def remove_directory_if_empty(dir_to_remove):
    """Remove the directory if it is empty, with one exception: the directory returned by
    FreeCAD.getUserMacroDir(True) will not be removed even if it is empty."""

    if dir_to_remove == fci.DataPaths().macro_dir:
        return
    if not os.listdir(dir_to_remove):
        os.rmdir(dir_to_remove)


def restart_freecad():
    """Shuts down and restarts FreeCAD"""

    if not QtCore or not QtWidgets:
        return

    args = QtWidgets.QApplication.arguments()[1:]
    if fci.FreeCADGui.getMainWindow().close():
        QtCore.QProcess.startDetached(QtWidgets.QApplication.applicationFilePath(), args)


def get_zip_url(repo):
    """Returns the location of a zip file from a repo, if available"""

    parsed_url = urlparse(repo.url)
    if parsed_url.netloc == "github.com":
        return f"{repo.url}/archive/{repo.branch}.zip"
    if parsed_url.netloc in ["gitlab.com", "framagit.org", "salsa.debian.org"]:
        return f"{repo.url}/-/archive/{repo.branch}/{repo.name}-{repo.branch}.zip"
    fci.Console.PrintLog(
        "Debug: addonmanager_utilities.get_zip_url: Unknown git host fetching zip URL:"
        + parsed_url.netloc
        + "\n"
    )
    return f"{repo.url}/-/archive/{repo.branch}/{repo.name}-{repo.branch}.zip"


def recognized_git_location(repo) -> bool:
    """Returns whether this repo is based at a known git repo location: works with github, gitlab,
    framagit, and salsa.debian.org"""

    parsed_url = urlparse(repo.url)
    return parsed_url.netloc in [
        "github.com",
        "gitlab.com",
        "framagit.org",
        "salsa.debian.org",
    ]


def construct_git_url(repo, filename):
    """Returns a direct download link to a file in an online Git repo"""

    parsed_url = urlparse(repo.url)
    if parsed_url.netloc == "github.com":
        return f"{repo.url}/raw/{repo.branch}/{filename}"
    if parsed_url.netloc in ["gitlab.com", "framagit.org", "salsa.debian.org"]:
        return f"{repo.url}/-/raw/{repo.branch}/{filename}"
    fci.Console.PrintLog(
        "Debug: addonmanager_utilities.construct_git_url: Unknown git host:"
        + parsed_url.netloc
        + f" for file {filename}\n"
    )
    # Assume it's some kind of local GitLab instance...
    return f"{repo.url}/-/raw/{repo.branch}/{filename}"


def get_readme_url(repo):
    """Returns the location of a readme file"""

    return construct_git_url(repo, "README.md")


def get_metadata_url(url):
    """Returns the location of a package.xml metadata file"""

    return construct_git_url(url, "package.xml")


def get_desc_regex(repo):
    """Returns a regex string that extracts a WB description to be displayed in the description
    panel of the Addon manager, if the README could not be found"""

    parsed_url = urlparse(repo.url)
    if parsed_url.netloc == "github.com":
        return r'<meta property="og:description" content="(.*?)"'
    if parsed_url.netloc in ["gitlab.com", "salsa.debian.org", "framagit.org"]:
        return r'<meta.*?content="(.*?)".*?og:description.*?>'
    fci.Console.PrintLog(
        "Debug: addonmanager_utilities.get_desc_regex: Unknown git host:",
        repo.url,
        "\n",
    )
    return r'<meta.*?content="(.*?)".*?og:description.*?>'


def get_readme_html_url(repo):
    """Returns the location of a html file containing readme"""

    parsed_url = urlparse(repo.url)
    if parsed_url.netloc == "github.com":
        return f"{repo.url}/blob/{repo.branch}/README.md"
    if parsed_url.netloc in ["gitlab.com", "salsa.debian.org", "framagit.org"]:
        return f"{repo.url}/-/blob/{repo.branch}/README.md"
    fci.Console.PrintLog("Unrecognized git repo location '' -- guessing it is a GitLab instance...")
    return f"{repo.url}/-/blob/{repo.branch}/README.md"


def is_darkmode() -> bool:
    """Heuristics to determine if we are in a darkmode stylesheet"""
    pl = fci.FreeCADGui.getMainWindow().palette()
    return pl.color(pl.Background).lightness() < 128


def warning_color_string() -> str:
    """A shade of red, adapted to darkmode if possible. Targets a minimum 7:1 contrast ratio."""
    return "rgb(255,105,97)" if is_darkmode() else "rgb(215,0,21)"


def bright_color_string() -> str:
    """A shade of green, adapted to darkmode if possible. Targets a minimum 7:1 contrast ratio."""
    return "rgb(48,219,91)" if is_darkmode() else "rgb(36,138,61)"


def attention_color_string() -> str:
    """A shade of orange, adapted to darkmode if possible. Targets a minimum 7:1 contrast ratio."""
    return "rgb(255,179,64)" if is_darkmode() else "rgb(255,149,0)"


def get_assigned_string_literal(line: str) -> Optional[str]:
    """Look for a line of the form my_var = "A string literal" and return the string literal.
    If the assignment is of a floating point value, that value is converted to a string
    and returned. If neither is true, returns None."""

    string_search_regex = re.compile(r"\s*(['\"])(.*)\1")
    _, _, after_equals = line.partition("=")
    match = re.match(string_search_regex, after_equals)
    if match:
        return str(match.group(2))
    if is_float(after_equals):
        return str(after_equals).strip()
    return None


def get_macro_version_from_file(filename: str) -> str:
    """Get the version of the macro from a local macro file. Supports strings, ints, and floats,
    as well as a reference to __date__"""

    date = ""
    with open(filename, errors="ignore", encoding="utf-8") as f:
        line_counter = 0
        max_lines_to_scan = 200
        while line_counter < max_lines_to_scan:
            line_counter += 1
            line = f.readline()
            if not line:  # EOF
                break
            if line.lower().startswith("__version__"):
                match = get_assigned_string_literal(line)
                if match:
                    return match
                if "__date__" in line.lower():
                    # Don't do any real syntax checking, just assume the line is something
                    # like __version__ = __date__
                    if date:
                        return date
                    # pylint: disable=line-too-long,consider-using-f-string
                    fci.Console.PrintWarning(
                        translate(
                            "AddonsInstaller",
                            "Macro {} specified '__version__ = __date__' prior to setting a value for __date__".format(
                                filename
                            ),
                        )
                    )
            elif line.lower().startswith("__date__"):
                match = get_assigned_string_literal(line)
                if match:
                    date = match
    return ""


def update_macro_installation_details(repo) -> None:
    """Determine if a given macro is installed, either in its plain name,
    or prefixed with "Macro_" """
    if repo is None or not hasattr(repo, "macro") or repo.macro is None:
        fci.Console.PrintLog("Requested macro details for non-macro object\n")
        return
    test_file_one = os.path.join(fci.DataPaths().macro_dir, repo.macro.filename)
    test_file_two = os.path.join(fci.DataPaths().macro_dir, "Macro_" + repo.macro.filename)
    if os.path.exists(test_file_one):
        repo.updated_timestamp = os.path.getmtime(test_file_one)
        repo.installed_version = get_macro_version_from_file(test_file_one)
    elif os.path.exists(test_file_two):
        repo.updated_timestamp = os.path.getmtime(test_file_two)
        repo.installed_version = get_macro_version_from_file(test_file_two)
    else:
        return


# Borrowed from Stack Overflow:
# https://stackoverflow.com/questions/736043/checking-if-a-string-can-be-converted-to-float
def is_float(element: Any) -> bool:
    """Determine whether a given item can be converted to a floating-point number"""
    try:
        float(element)
        return True
    except ValueError:
        return False


def get_python_exe() -> str:
    """Find Python. In preference order
    A) The value of the PythonExecutableForPip user preference
    B) The executable located in the same bin directory as FreeCAD and called "python3"
    C) The executable located in the same bin directory as FreeCAD and called "python"
    D) The result of a shutil search for your system's "python3" executable
    E) The result of a shutil search for your system's "python" executable"""
    prefs = fci.ParamGet("User parameter:BaseApp/Preferences/Addons")
    python_exe = prefs.GetString("PythonExecutableForPip", "Not set")
    fc_dir = fci.DataPaths().home_dir
    if not python_exe or python_exe == "Not set" or not os.path.exists(python_exe):
        python_exe = os.path.join(fc_dir, "bin", "python3")
        if "Windows" in platform.system():
            python_exe += ".exe"

    if not python_exe or not os.path.exists(python_exe):
        python_exe = os.path.join(fc_dir, "bin", "python")
        if "Windows" in platform.system():
            python_exe += ".exe"

    if not python_exe or not os.path.exists(python_exe):
        python_exe = shutil.which("python3")

    if not python_exe or not os.path.exists(python_exe):
        python_exe = shutil.which("python")

    if not python_exe or not os.path.exists(python_exe):
        return ""

    python_exe = python_exe.replace("/", os.path.sep)
    prefs.SetString("PythonExecutableForPip", python_exe)
    return python_exe


def get_pip_target_directory():
    # Get the default location to install new pip packages
    major, minor, _ = platform.python_version_tuple()
    vendor_path = os.path.join(
        fci.DataPaths().mod_dir, "..", "AdditionalPythonPackages", f"py{major}{minor}"
    )
    return vendor_path


def get_cache_file_name(file: str) -> str:
    """Get the full path to a cache file with a given name."""
    cache_path = fci.DataPaths().cache_dir
    am_path = os.path.join(cache_path, "AddonManager")
    os.makedirs(am_path, exist_ok=True)
    return os.path.join(am_path, file)


def blocking_get(url: str, method=None) -> bytes:
    """Wrapper around three possible ways of accessing data, depending on the current run mode and
    Python installation. Blocks until complete, and returns the text results of the call if it
    succeeded, or an empty string if it failed, or returned no data. The method argument is
    provided mainly for testing purposes."""
    p = b""
    if fci.FreeCADGui and method is None or method == "networkmanager":
        NetworkManager.InitializeNetworkManager()
        p = NetworkManager.AM_NETWORK_MANAGER.blocking_get(url)
        if p:
            try:
                p = p.data()
            except AttributeError:
                pass
    elif requests and method is None or method == "requests":
        response = requests.get(url)
        if response.status_code == 200:
            p = response.content
    else:
        ctx = ssl.create_default_context()
        with urllib.request.urlopen(url, context=ctx) as f:
            p = f.read()
    return p


def run_interruptable_subprocess(args) -> subprocess.CompletedProcess:
    """Wrap subprocess call so it can be interrupted gracefully."""
    creation_flags = 0
    if hasattr(subprocess, "CREATE_NO_WINDOW"):
        # Added in Python 3.7 -- only used on Windows
        creation_flags = subprocess.CREATE_NO_WINDOW
    try:
        p = subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            creationflags=creation_flags,
            text=True,
            encoding="utf-8",
        )
    except OSError as e:
        raise subprocess.CalledProcessError(-1, args, "", e.strerror)
    stdout = ""
    stderr = ""
    return_code = None
    while return_code is None:
        try:
            stdout, stderr = p.communicate(timeout=10)
            return_code = p.returncode
        except subprocess.TimeoutExpired:
            if QtCore.QThread.currentThread().isInterruptionRequested():
                p.kill()
                raise ProcessInterrupted()
    if return_code is None or return_code != 0:
        raise subprocess.CalledProcessError(
            return_code if return_code is not None else -1, args, stdout, stderr
        )
    return subprocess.CompletedProcess(args, return_code, stdout, stderr)


def get_main_am_window():
    windows = QtWidgets.QApplication.topLevelWidgets()
    for widget in windows:
        if widget.objectName() == "AddonManager_Main_Window":
            return widget
    # If there is no main AM window, we may be running unit tests: see if the Test Runner window
    # exists:
    for widget in windows:
        if widget.objectName() == "TestGui__UnitTest":
            return widget
    # If we still didn't find it, try to locate the main FreeCAD window:
    for widget in windows:
        if hasattr(widget, "centralWidget"):
            return widget.centralWidget()
    # Why is this code even getting called?
    return None



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint displacement"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import FreeCAD


def get_analysis_types():
    return "all"    # write for all analysis types


def get_sets_name():
    return "constraints_displacement_node_sets"


def get_constraint_title():
    return "Displacement constraint applied"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return "\n"


def write_meshdata_constraint(f, femobj, disp_obj, ccxwriter):
    f.write("*NSET,NSET={}\n".format(disp_obj.Name))
    for n in femobj["Nodes"]:
        f.write("{},\n".format(n))


def write_constraint(f, femobj, disp_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    f.write("*BOUNDARY\n")
    if disp_obj.xFix:
        f.write("{},1\n".format(disp_obj.Name))
    elif not disp_obj.xFree:
        f.write(
            "{},1,1,{}\n".format(
                disp_obj.Name, FreeCAD.Units.Quantity(disp_obj.xDisplacement.getValueAs("mm"))
            )
        )
    if disp_obj.yFix:
        f.write("{},2\n".format(disp_obj.Name))
    elif not disp_obj.yFree:
        f.write(
            "{},2,2,{}\n".format(
                disp_obj.Name, FreeCAD.Units.Quantity(disp_obj.yDisplacement.getValueAs("mm"))
            )
        )
    if disp_obj.zFix:
        f.write("{},3\n".format(disp_obj.Name))
    elif not disp_obj.zFree:
        f.write(
            "{},3,3,{}\n".format(
                disp_obj.Name, FreeCAD.Units.Quantity(disp_obj.zDisplacement.getValueAs("mm"))
            )
        )

    if ccxwriter.member.geos_beamsection or ccxwriter.member.geos_shellthickness:
        if disp_obj.rotxFix:
            f.write("{},4\n".format(disp_obj.Name))
        elif not disp_obj.rotxFree:
            f.write(
                "{},4,4,{}\n".format(
                    disp_obj.Name, FreeCAD.Units.Quantity(disp_obj.xRotation.getValueAs("deg"))
                )
            )
        if disp_obj.rotyFix:
            f.write("{},5\n".format(disp_obj.Name))
        elif not disp_obj.rotyFree:
            f.write(
                "{},5,5,{}\n".format(
                    disp_obj.Name, FreeCAD.Units.Quantity(disp_obj.yRotation.getValueAs("deg"))
                )
            )
        if disp_obj.rotzFix:
            f.write("{},6\n".format(disp_obj.Name))
        elif not disp_obj.rotzFree:
            f.write(
                "{},6,6,{}\n".format(
                    disp_obj.Name, FreeCAD.Units.Quantity(disp_obj.zRotation.getValueAs("deg"))
                )
            )



# -*- coding: utf-8 -*-
# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import TestApp

from PathTests.TestPathProfile import TestPathProfile

from PathTests.TestPathAdaptive import TestPathAdaptive
from PathTests.TestPathCore import TestPathCore
from PathTests.TestPathDepthParams import depthTestCases
from PathTests.TestPathDressupDogbone import TestDressupDogbone
from PathTests.TestPathDressupDogboneII import TestDressupDogboneII
from PathTests.TestPathDressupHoldingTags import TestHoldingTags
from PathTests.TestPathDrillable import TestPathDrillable
from PathTests.TestPathDrillGenerator import TestPathDrillGenerator
from PathTests.TestPathGeneratorDogboneII import TestGeneratorDogboneII
from PathTests.TestPathGeom import TestPathGeom
from PathTests.TestPathLanguage import TestPathLanguage
from PathTests.TestPathOpDeburr import TestPathOpDeburr

# from PathTests.TestPathHelix import TestPathHelix
from PathTests.TestPathHelpers import TestPathHelpers
from PathTests.TestPathHelixGenerator import TestPathHelixGenerator
from PathTests.TestPathLog import TestPathLog
from PathTests.TestPathOpUtil import TestPathOpUtil

# from PathTests.TestPathPost import TestPathPost
from PathTests.TestPathPost import TestPathPostUtils
from PathTests.TestPathPost import TestBuildPostList
from PathTests.TestPathPost import TestOutputNameSubstitution

from PathTests.TestPathPreferences import TestPathPreferences
from PathTests.TestPathProfile import TestPathProfile
from PathTests.TestPathPropertyBag import TestPathPropertyBag
from PathTests.TestPathRotationGenerator import TestPathRotationGenerator
from PathTests.TestPathSetupSheet import TestPathSetupSheet
from PathTests.TestPathStock import TestPathStock
from PathTests.TestPathThreadMilling import TestPathThreadMilling
from PathTests.TestPathThreadMillingGenerator import TestPathThreadMillingGenerator
from PathTests.TestPathToolBit import TestPathToolBit
from PathTests.TestPathToolChangeGenerator import TestPathToolChangeGenerator
from PathTests.TestPathToolController import TestPathToolController
from PathTests.TestPathUtil import TestPathUtil
from PathTests.TestPathVcarve import TestPathVcarve
from PathTests.TestPathVoronoi import TestPathVoronoi

from PathTests.TestCentroidPost import TestCentroidPost
from PathTests.TestGrblPost import TestGrblPost
from PathTests.TestLinuxCNCPost import TestLinuxCNCPost
from PathTests.TestMach3Mach4Post import TestMach3Mach4Post
from PathTests.TestRefactoredCentroidPost import TestRefactoredCentroidPost
from PathTests.TestRefactoredGrblPost import TestRefactoredGrblPost
from PathTests.TestRefactoredLinuxCNCPost import TestRefactoredLinuxCNCPost
from PathTests.TestRefactoredMach3Mach4Post import TestRefactoredMach3Mach4Post
from PathTests.TestRefactoredTestPost import TestRefactoredTestPost

# dummy usage to get flake8 and lgtm quiet
False if depthTestCases.__name__ else True
False if TestApp.__name__ else True
False if TestBuildPostList.__name__ else True
False if TestDressupDogbone.__name__ else True
False if TestDressupDogboneII.__name__ else True
False if TestGeneratorDogboneII.__name__ else True
False if TestHoldingTags.__name__ else True
False if TestPathLanguage.__name__ else True
False if TestOutputNameSubstitution.__name__ else True
False if TestPathAdaptive.__name__ else True
False if TestPathCore.__name__ else True
False if TestPathOpDeburr.__name__ else True
False if TestPathDrillable.__name__ else True
False if TestPathGeom.__name__ else True
False if TestPathHelpers.__name__ else True
# False if TestPathHelix.__name__ else True
False if TestPathLog.__name__ else True
False if TestPathOpUtil.__name__ else True
# False if TestPathPost.__name__ else True
False if TestPathPostUtils.__name__ else True
False if TestPathPreferences.__name__ else True
False if TestPathProfile.__name__ else True
False if TestPathPropertyBag.__name__ else True
False if TestPathRotationGenerator.__name__ else True
False if TestPathSetupSheet.__name__ else True
False if TestPathStock.__name__ else True
False if TestPathThreadMilling.__name__ else True
False if TestPathThreadMillingGenerator.__name__ else True
False if TestPathToolBit.__name__ else True
False if TestPathToolChangeGenerator.__name__ else True
False if TestPathToolController.__name__ else True
False if TestPathUtil.__name__ else True
False if TestPathVcarve.__name__ else True
False if TestPathVoronoi.__name__ else True
False if TestPathDrillGenerator.__name__ else True
False if TestPathHelixGenerator.__name__ else True

False if TestCentroidPost.__name__ else True
False if TestGrblPost.__name__ else True
False if TestLinuxCNCPost.__name__ else True
False if TestMach3Mach4Post.__name__ else True
False if TestRefactoredCentroidPost.__name__ else True
False if TestRefactoredGrblPost.__name__ else True
False if TestRefactoredLinuxCNCPost.__name__ else True
False if TestRefactoredMach3Mach4Post.__name__ else True
False if TestRefactoredTestPost.__name__ else True



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

""" Contains a class to handle editing a person (from a Metadata standpoint). """

import os
from typing import Tuple  # Needed until Py 3.9, when tuple supports this directly

from PySide.QtWidgets import QDialog

import FreeCAD
import FreeCADGui

translate = FreeCAD.Qt.translate


class PersonEditor:
    """Create or edit a maintainer or author record."""

    def __init__(self):

        self.dialog = FreeCADGui.PySideUic.loadUi(
            os.path.join(os.path.dirname(__file__), "developer_mode_people.ui")
        )
        self.dialog.comboBox.clear()
        self.dialog.comboBox.addItem(
            translate("AddonsInstaller", "Maintainer"), userData="maintainer"
        )
        self.dialog.comboBox.addItem(
            translate("AddonsInstaller", "Author"), userData="author"
        )

    def exec(self) -> Tuple[str, str, str]:
        """Run the dialog, and return a tuple of the person's record type, their name, and their
        email address. Email may be None. If the others are None it's because the user cancelled
        the interaction."""
        result = self.dialog.exec()
        if result == QDialog.Accepted:
            return (
                self.dialog.comboBox.currentData(),
                self.dialog.nameLineEdit.text(),
                self.dialog.emailLineEdit.text(),
            )
        return "", "", ""

    def setup(
        self, person_type: str = "maintainer", name: str = "", email: str = ""
    ) -> None:
        """Configure the dialog"""
        index = self.dialog.comboBox.findData(person_type)
        if index == -1:
            FreeCAD.Console.PrintWarning(
                f"Internal Error: unrecognized person type {person_type}"
            )
            index = 0
        self.dialog.comboBox.setCurrentIndex(index)
        self.dialog.nameLineEdit.setText(name)
        self.dialog.emailLineEdit.setText(email)



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2022 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import os
import tempfile
import unittest
import FreeCAD

from PySide import QtCore, QtWidgets

from addonmanager_installer_gui import AddonInstallerGUI, MacroInstallerGUI

from AddonManagerTest.gui.gui_mocks import DialogWatcher, DialogInteractor
from AddonManagerTest.app.mocks import MockAddon

translate = FreeCAD.Qt.translate


class TestInstallerGui(unittest.TestCase):

    MODULE = "test_installer_gui"  # file name without extension

    def setUp(self):
        self.addon_to_install = MockAddon()
        self.installer_gui = AddonInstallerGUI(self.addon_to_install)
        self.finalized_thread = False

    def tearDown(self):
        pass

    def test_success_dialog(self):
        # Pop the modal dialog and verify that it opens, and responds to an OK click
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Success"),
            QtWidgets.QDialogButtonBox.Ok,
        )
        self.installer_gui._installation_succeeded()
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_failure_dialog(self):
        # Pop the modal dialog and verify that it opens, and responds to a Cancel click
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Installation Failed"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        self.installer_gui._installation_failed(
            self.addon_to_install, "Test of installation failure"
        )
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_no_python_dialog(self):
        # Pop the modal dialog and verify that it opens, and responds to a No click
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Cannot execute Python"),
            QtWidgets.QDialogButtonBox.No,
        )
        self.installer_gui._report_no_python_exe()
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_no_pip_dialog(self):
        # Pop the modal dialog and verify that it opens, and responds to a No click
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Cannot execute pip"),
            QtWidgets.QDialogButtonBox.No,
        )
        self.installer_gui._report_no_pip("pip not actually run, this was a test")
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_dependency_failure_dialog(self):
        # Pop the modal dialog and verify that it opens, and responds to a No click
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Package installation failed"),
            QtWidgets.QDialogButtonBox.No,
        )
        self.installer_gui._report_dependency_failure(
            "Unit test", "Nothing really failed, this is a test of the dialog box"
        )
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_install(self):
        # Run the installation code and make sure it puts the directory in place
        with tempfile.TemporaryDirectory() as temp_dir:
            self.installer_gui.installer.installation_path = temp_dir
            self.installer_gui.install()  # This does not block
            self.installer_gui.installer.success.disconnect(
                self.installer_gui._installation_succeeded
            )
            self.installer_gui.installer.failure.disconnect(
                self.installer_gui._installation_failed
            )
            while not self.installer_gui.worker_thread.isFinished():
                QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents, 100)
            self.assertTrue(
                os.path.exists(os.path.join(temp_dir, "MockAddon")),
                "Installed directory not found",
            )

    def test_handle_disallowed_python(self):
        disallowed_packages = ["disallowed_package_name"]
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Missing Requirement"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        self.installer_gui._handle_disallowed_python(disallowed_packages)
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_handle_disallowed_python_long_list(self):
        """A separate test for when there are MANY packages, which takes a separate code path."""
        disallowed_packages = []
        for i in range(50):
            disallowed_packages.append(f"disallowed_package_name_{i}")
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Missing Requirement"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        self.installer_gui._handle_disallowed_python(disallowed_packages)
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_report_missing_workbenches_single(self):
        """Test only missing one workbench"""
        wbs = ["OneMissingWorkbench"]
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Missing Requirement"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        self.installer_gui._report_missing_workbenches(wbs)
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_report_missing_workbenches_multiple(self):
        """Test only missing one workbench"""
        wbs = ["FirstMissingWorkbench", "SecondMissingWorkbench"]
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Missing Requirement"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        self.installer_gui._report_missing_workbenches(wbs)
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_resolve_dependencies_then_install(self):
        class MissingDependenciesMock:
            def __init__(self):
                self.external_addons = ["addon_1", "addon_2"]
                self.python_requires = ["py_req_1", "py_req_2"]
                self.python_optional = ["py_opt_1", "py_opt_2"]

        missing = MissingDependenciesMock()
        dialog_watcher = DialogWatcher(
            translate("DependencyResolutionDialog", "Resolve Dependencies"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        self.installer_gui._resolve_dependencies_then_install(missing)
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_check_python_version_bad(self):
        class MissingDependenciesMock:
            def __init__(self):
                self.python_min_version = {"major": 3, "minor": 9999}

        missing = MissingDependenciesMock()
        dialog_watcher = DialogWatcher(
            translate("AddonsInstaller", "Incompatible Python version"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        stop_installing = self.installer_gui._check_python_version(missing)
        self.assertTrue(
            dialog_watcher.dialog_found, "Failed to find the expected dialog box"
        )
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )
        self.assertTrue(
            stop_installing, "Failed to halt installation on bad Python version"
        )

    def test_check_python_version_good(self):
        class MissingDependenciesMock:
            def __init__(self):
                self.python_min_version = {"major": 3, "minor": 0}

        missing = MissingDependenciesMock()
        stop_installing = self.installer_gui._check_python_version(missing)
        self.assertFalse(
            stop_installing, "Failed to continue installation on good Python version"
        )

    def test_clean_up_optional(self):
        class MissingDependenciesMock:
            def __init__(self):
                self.python_optional = [
                    "allowed_packages_1",
                    "allowed_packages_2",
                    "disallowed_package",
                ]

        allowed_packages = ["allowed_packages_1", "allowed_packages_2"]
        missing = MissingDependenciesMock()
        self.installer_gui.installer.allowed_packages = set(allowed_packages)
        self.installer_gui._clean_up_optional(missing)
        self.assertTrue("allowed_packages_1" in missing.python_optional)
        self.assertTrue("allowed_packages_2" in missing.python_optional)
        self.assertFalse("disallowed_package" in missing.python_optional)

    def intercept_run_dependency_installer(
        self, addons, python_requires, python_optional
    ):
        self.assertEqual(python_requires, ["py_req_1", "py_req_2"])
        self.assertEqual(python_optional, ["py_opt_1", "py_opt_2"])
        self.assertEqual(addons[0].name, "addon_1")
        self.assertEqual(addons[1].name, "addon_2")

    def test_dependency_dialog_yes_clicked(self):
        class DialogMock:
            class ListWidgetMock:
                class ListWidgetItemMock:
                    def __init__(self, name):
                        self.name = name

                    def text(self):
                        return self.name

                    def checkState(self):
                        return QtCore.Qt.Checked

                def __init__(self, items):
                    self.list = []
                    for item in items:
                        self.list.append(
                            DialogMock.ListWidgetMock.ListWidgetItemMock(item)
                        )

                def count(self):
                    return len(self.list)

                def item(self, i):
                    return self.list[i]

            def __init__(self):
                self.listWidgetAddons = DialogMock.ListWidgetMock(
                    ["addon_1", "addon_2"]
                )
                self.listWidgetPythonRequired = DialogMock.ListWidgetMock(
                    ["py_req_1", "py_req_2"]
                )
                self.listWidgetPythonOptional = DialogMock.ListWidgetMock(
                    ["py_opt_1", "py_opt_2"]
                )

        class AddonMock:
            def __init__(self, name):
                self.name = name

        self.installer_gui.dependency_dialog = DialogMock()
        self.installer_gui.addons = [AddonMock("addon_1"), AddonMock("addon_2")]
        self.installer_gui._run_dependency_installer = (
            self.intercept_run_dependency_installer
        )
        self.installer_gui._dependency_dialog_yes_clicked()


class TestMacroInstallerGui(unittest.TestCase):
    class MockMacroAddon:
        class MockMacro:
            def __init__(self):
                self.install_called = False
                self.install_result = (
                    True  # External code can change to False to test failed install
                )
                self.name = "MockMacro"
                self.filename = "mock_macro_no_real_file.FCMacro"
                self.comment = "This is a mock macro for unit testing"
                self.icon = None
                self.xpm = None

            def install(self):
                self.install_called = True
                return self.install_result

        def __init__(self):
            self.macro = TestMacroInstallerGui.MockMacroAddon.MockMacro()
            self.name = self.macro.name
            self.display_name = self.macro.name

    class MockParameter:
        """Mock the parameter group to allow simplified behavior and introspection."""

        def __init__(self):
            self.params = {}
            self.groups = {}
            self.accessed_parameters = {}  # Dict is param name: default value

            types = ["Bool", "String", "Int", "UInt", "Float"]
            for t in types:
                setattr(self, f"Get{t}", self.get)
                setattr(self, f"Set{t}", self.set)
                setattr(self, f"Rem{t}", self.rem)

        def get(self, p, default=None):
            self.accessed_parameters[p] = default
            if p in self.params:
                return self.params[p]
            else:
                return default

        def set(self, p, value):
            self.params[p] = value

        def rem(self, p):
            if p in self.params:
                self.params.erase(p)

        def GetGroup(self, name):
            if name not in self.groups:
                self.groups[name] = TestMacroInstallerGui.MockParameter()
            return self.groups[name]

        def GetGroups(self):
            return self.groups.keys()

    class ToolbarIntercepter:
        def __init__(self):
            self.ask_for_toolbar_called = False
            self.install_macro_to_toolbar_called = False
            self.tb = None
            self.custom_group = TestMacroInstallerGui.MockParameter()
            self.custom_group.set("Name", "MockCustomToolbar")

        def _ask_for_toolbar(self, _):
            self.ask_for_toolbar_called = True
            return self.custom_group

        def _install_macro_to_toolbar(self, tb):
            self.install_macro_to_toolbar_called = True
            self.tb = tb

    class InstallerInterceptor:
        def __init__(self):
            self.ccc_called = False

        def _create_custom_command(
            self,
            toolbar,
            filename,
            menuText,
            tooltipText,
            whatsThisText,
            statustipText,
            pixmapText,
        ):
            self.ccc_called = True
            self.toolbar = toolbar
            self.filename = filename
            self.menuText = menuText
            self.tooltipText = tooltipText
            self.whatsThisText = whatsThisText
            self.statustipText = statustipText
            self.pixmapText = pixmapText

    def setUp(self):
        self.mock_macro = TestMacroInstallerGui.MockMacroAddon()
        self.installer = MacroInstallerGUI(self.mock_macro)
        self.installer.addon_params = TestMacroInstallerGui.MockParameter()
        self.installer.toolbar_params = TestMacroInstallerGui.MockParameter()

    def tearDown(self):
        pass

    def test_class_is_initialized(self):
        """Connecting to a signal does not throw"""
        self.installer.finished.connect(lambda: None)

    def test_ask_for_toolbar_no_dialog_default_exists(self):
        self.installer.addon_params.set("alwaysAskForToolbar", False)
        self.installer.addon_params.set("CustomToolbarName", "UnitTestCustomToolbar")
        utct = self.installer.toolbar_params.GetGroup("UnitTestCustomToolbar")
        utct.set("Name", "UnitTestCustomToolbar")
        utct.set("Active", True)
        result = self.installer._ask_for_toolbar([])
        self.assertIsNotNone(result)
        self.assertTrue(hasattr(result, "get"))
        name = result.get("Name")
        self.assertEqual(name, "UnitTestCustomToolbar")

    def test_ask_for_toolbar_with_dialog_cancelled(self):

        # First test: the user cancels the dialog
        self.installer.addon_params.set("alwaysAskForToolbar", True)
        dialog_watcher = DialogWatcher(
            translate("select_toolbar_dialog", "Select Toolbar"),
            QtWidgets.QDialogButtonBox.Cancel,
        )
        result = self.installer._ask_for_toolbar([])
        self.assertIsNone(result)

    def test_ask_for_toolbar_with_dialog_defaults(self):

        # Second test: the user leaves the dialog at all default values, so:
        #   - The checkbox "Ask every time" is unchecked
        #   - The selected toolbar option is "Create new toolbar", which triggers a search for
        # a new custom toolbar name by calling _create_new_custom_toolbar, which we mock.
        fake_custom_toolbar_group = TestMacroInstallerGui.MockParameter()
        fake_custom_toolbar_group.set("Name", "UnitTestCustomToolbar")
        self.installer._create_new_custom_toolbar = lambda: fake_custom_toolbar_group
        dialog_watcher = DialogWatcher(
            translate("select_toolbar_dialog", "Select Toolbar"),
            QtWidgets.QDialogButtonBox.Ok,
        )
        result = self.installer._ask_for_toolbar([])
        self.assertIsNotNone(result)
        self.assertTrue(hasattr(result, "get"))
        name = result.get("Name")
        self.assertEqual(name, "UnitTestCustomToolbar")
        self.assertIn("alwaysAskForToolbar", self.installer.addon_params.params)
        self.assertFalse(self.installer.addon_params.get("alwaysAskForToolbar", True))
        self.assertTrue(
            dialog_watcher.button_found, "Failed to find the expected button"
        )

    def test_ask_for_toolbar_with_dialog_selection(self):

        # Third test: the user selects a custom toolbar in the dialog, and checks the box to always
        # ask.
        dialog_interactor = DialogInteractor(
            translate("select_toolbar_dialog", "Select Toolbar"),
            self.interactor_selection_option_and_checkbox,
        )
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_2 = self.installer.toolbar_params.GetGroup("UT_TB_2")
        ut_tb_3 = self.installer.toolbar_params.GetGroup("UT_TB_3")
        ut_tb_1.set("Name", "UT_TB_1")
        ut_tb_2.set("Name", "UT_TB_2")
        ut_tb_3.set("Name", "UT_TB_3")
        result = self.installer._ask_for_toolbar(["UT_TB_1", "UT_TB_2", "UT_TB_3"])
        self.assertIsNotNone(result)
        self.assertTrue(hasattr(result, "get"))
        name = result.get("Name")
        self.assertEqual(name, "UT_TB_3")
        self.assertIn("alwaysAskForToolbar", self.installer.addon_params.params)
        self.assertTrue(self.installer.addon_params.get("alwaysAskForToolbar", False))

    def interactor_selection_option_and_checkbox(self, parent):

        boxes = parent.findChildren(QtWidgets.QComboBox)
        self.assertEqual(len(boxes), 1)  # Just to make sure...
        box = boxes[0]
        box.setCurrentIndex(box.count() - 2)  # Select the last thing but one

        checkboxes = parent.findChildren(QtWidgets.QCheckBox)
        self.assertEqual(len(checkboxes), 1)  # Just to make sure...
        checkbox = checkboxes[0]
        checkbox.setChecked(True)

        parent.accept()

    def test_macro_button_exists_no_command(self):
        # Test 1: No command for this macro
        self.installer._find_custom_command = lambda _: None
        button_exists = self.installer._macro_button_exists()
        self.assertFalse(button_exists)

    def test_macro_button_exists_true(self):
        # Test 2: Macro is in the list of buttons
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UnitTestCommand")
        ut_tb_1.set(
            "UnitTestCommand", "FreeCAD"
        )  # This is what the real thing looks like...
        self.installer._find_custom_command = lambda _: "UnitTestCommand"
        self.assertTrue(self.installer._macro_button_exists())

    def test_macro_button_exists_false(self):
        # Test 3: Macro is not in the list of buttons
        self.installer._find_custom_command = lambda _: "UnitTestCommand"
        self.assertFalse(self.installer._macro_button_exists())

    def test_ask_to_install_toolbar_button_disabled(self):
        self.installer.addon_params.SetBool("dontShowAddMacroButtonDialog", True)
        self.installer._ask_to_install_toolbar_button()
        # This should NOT block when dontShowAddMacroButtonDialog is True

    def test_ask_to_install_toolbar_button_enabled_no(self):
        self.installer.addon_params.SetBool("dontShowAddMacroButtonDialog", False)
        dialog_watcher = DialogWatcher(
            translate("toolbar_button", "Add button?"),
            QtWidgets.QDialogButtonBox.No,
        )
        # Note: that dialog does not use a QButtonBox, so we can really only test its
        # reject() signal, which is triggered by the DialogWatcher when it cannot find
        # the button. In this case, failure to find that button is NOT an error.
        self.installer._ask_to_install_toolbar_button()  # Blocks until killed by watcher
        self.assertTrue(dialog_watcher.dialog_found)

    def test_get_toolbar_with_name_found(self):
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UnitTestToolbar")
        ut_tb_1.set("Name", "Unit Test Toolbar")
        ut_tb_1.set("UnitTestParam", True)
        tb = self.installer._get_toolbar_with_name("Unit Test Toolbar")
        self.assertIsNotNone(tb)
        self.assertTrue(tb.get("UnitTestParam", False))

    def test_get_toolbar_with_name_not_found(self):
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UnitTestToolbar")
        ut_tb_1.set("Name", "Not the Unit Test Toolbar")
        tb = self.installer._get_toolbar_with_name("Unit Test Toolbar")
        self.assertIsNone(tb)

    def test_create_new_custom_toolbar_no_existing(self):
        tb = self.installer._create_new_custom_toolbar()
        self.assertEqual(tb.get("Name", ""), "Auto-Created Macro Toolbar")
        self.assertTrue(tb.get("Active", False), True)

    def test_create_new_custom_toolbar_one_existing(self):
        _ = self.installer._create_new_custom_toolbar()
        tb = self.installer._create_new_custom_toolbar()
        self.assertEqual(tb.get("Name", ""), "Auto-Created Macro Toolbar (2)")
        self.assertTrue(tb.get("Active", False), True)

    def test_check_for_toolbar_true(self):
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_1.set("Name", "UT_TB_1")
        self.assertTrue(self.installer._check_for_toolbar("UT_TB_1"))

    def test_check_for_toolbar_false(self):
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_1.set("Name", "UT_TB_1")
        self.assertFalse(self.installer._check_for_toolbar("Not UT_TB_1"))

    def test_install_toolbar_button_first_custom_toolbar(self):
        tbi = TestMacroInstallerGui.ToolbarIntercepter()
        self.installer._ask_for_toolbar = tbi._ask_for_toolbar
        self.installer._install_macro_to_toolbar = tbi._install_macro_to_toolbar
        self.installer._install_toolbar_button()
        self.assertTrue(tbi.install_macro_to_toolbar_called)
        self.assertFalse(tbi.ask_for_toolbar_called)
        self.assertTrue("Custom_1" in self.installer.toolbar_params.GetGroups())

    def test_install_toolbar_button_existing_custom_toolbar_1(self):
        # There is an existing custom toolbar, and we should use it
        tbi = TestMacroInstallerGui.ToolbarIntercepter()
        self.installer._ask_for_toolbar = tbi._ask_for_toolbar
        self.installer._install_macro_to_toolbar = tbi._install_macro_to_toolbar
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_1.set("Name", "UT_TB_1")
        self.installer.addon_params.set("CustomToolbarName", "UT_TB_1")
        self.installer._install_toolbar_button()
        self.assertTrue(tbi.install_macro_to_toolbar_called)
        self.assertFalse(tbi.ask_for_toolbar_called)
        self.assertEqual(tbi.tb.get("Name", ""), "UT_TB_1")

    def test_install_toolbar_button_existing_custom_toolbar_2(self):
        # There are multiple existing custom toolbars, and we should use one of them
        tbi = TestMacroInstallerGui.ToolbarIntercepter()
        self.installer._ask_for_toolbar = tbi._ask_for_toolbar
        self.installer._install_macro_to_toolbar = tbi._install_macro_to_toolbar
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_2 = self.installer.toolbar_params.GetGroup("UT_TB_2")
        ut_tb_3 = self.installer.toolbar_params.GetGroup("UT_TB_3")
        ut_tb_1.set("Name", "UT_TB_1")
        ut_tb_2.set("Name", "UT_TB_2")
        ut_tb_3.set("Name", "UT_TB_3")
        self.installer.addon_params.set("CustomToolbarName", "UT_TB_3")
        self.installer._install_toolbar_button()
        self.assertTrue(tbi.install_macro_to_toolbar_called)
        self.assertFalse(tbi.ask_for_toolbar_called)
        self.assertEqual(tbi.tb.get("Name", ""), "UT_TB_3")

    def test_install_toolbar_button_existing_custom_toolbar_3(self):
        # There are multiple existing custom toolbars, but none of them match
        tbi = TestMacroInstallerGui.ToolbarIntercepter()
        self.installer._ask_for_toolbar = tbi._ask_for_toolbar
        self.installer._install_macro_to_toolbar = tbi._install_macro_to_toolbar
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_2 = self.installer.toolbar_params.GetGroup("UT_TB_2")
        ut_tb_3 = self.installer.toolbar_params.GetGroup("UT_TB_3")
        ut_tb_1.set("Name", "UT_TB_1")
        ut_tb_2.set("Name", "UT_TB_2")
        ut_tb_3.set("Name", "UT_TB_3")
        self.installer.addon_params.set("CustomToolbarName", "UT_TB_4")
        self.installer._install_toolbar_button()
        self.assertTrue(tbi.install_macro_to_toolbar_called)
        self.assertTrue(tbi.ask_for_toolbar_called)
        self.assertEqual(tbi.tb.get("Name", ""), "MockCustomToolbar")

    def test_install_toolbar_button_existing_custom_toolbar_4(self):
        # There are multiple existing custom toolbars, one of them matches, but we have set
        # "alwaysAskForToolbar" to True
        tbi = TestMacroInstallerGui.ToolbarIntercepter()
        self.installer._ask_for_toolbar = tbi._ask_for_toolbar
        self.installer._install_macro_to_toolbar = tbi._install_macro_to_toolbar
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_2 = self.installer.toolbar_params.GetGroup("UT_TB_2")
        ut_tb_3 = self.installer.toolbar_params.GetGroup("UT_TB_3")
        ut_tb_1.set("Name", "UT_TB_1")
        ut_tb_2.set("Name", "UT_TB_2")
        ut_tb_3.set("Name", "UT_TB_3")
        self.installer.addon_params.set("CustomToolbarName", "UT_TB_3")
        self.installer.addon_params.set("alwaysAskForToolbar", True)
        self.installer._install_toolbar_button()
        self.assertTrue(tbi.install_macro_to_toolbar_called)
        self.assertTrue(tbi.ask_for_toolbar_called)
        self.assertEqual(tbi.tb.get("Name", ""), "MockCustomToolbar")

    def test_install_macro_to_toolbar_icon_abspath(self):
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_1.set("Name", "UT_TB_1")
        ii = TestMacroInstallerGui.InstallerInterceptor()
        self.installer._create_custom_command = ii._create_custom_command
        with tempfile.NamedTemporaryFile() as ntf:
            self.mock_macro.macro.icon = ntf.name
            self.installer._install_macro_to_toolbar(ut_tb_1)
            self.assertTrue(ii.ccc_called)
            self.assertEqual(ii.pixmapText, ntf.name)

    def test_install_macro_to_toolbar_icon_relpath(self):
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_1.set("Name", "UT_TB_1")
        ii = TestMacroInstallerGui.InstallerInterceptor()
        self.installer._create_custom_command = ii._create_custom_command
        with tempfile.TemporaryDirectory() as td:
            self.installer.macro_dir = td
            self.mock_macro.macro.icon = "RelativeIconPath.png"
            self.installer._install_macro_to_toolbar(ut_tb_1)
            self.assertTrue(ii.ccc_called)
            self.assertEqual(ii.pixmapText, os.path.join(td, "RelativeIconPath.png"))

    def test_install_macro_to_toolbar_xpm(self):
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_1.set("Name", "UT_TB_1")
        ii = TestMacroInstallerGui.InstallerInterceptor()
        self.installer._create_custom_command = ii._create_custom_command
        with tempfile.TemporaryDirectory() as td:
            self.installer.macro_dir = td
            self.mock_macro.macro.xpm = "Not really xpm data, don't try to use it!"
            self.installer._install_macro_to_toolbar(ut_tb_1)
            self.assertTrue(ii.ccc_called)
            self.assertEqual(ii.pixmapText, os.path.join(td, "MockMacro_icon.xpm"))
            self.assertTrue(os.path.exists(os.path.join(td, "MockMacro_icon.xpm")))

    def test_install_macro_to_toolbar_no_icon(self):
        ut_tb_1 = self.installer.toolbar_params.GetGroup("UT_TB_1")
        ut_tb_1.set("Name", "UT_TB_1")
        ii = TestMacroInstallerGui.InstallerInterceptor()
        self.installer._create_custom_command = ii._create_custom_command
        with tempfile.TemporaryDirectory() as td:
            self.installer.macro_dir = td
            self.installer._install_macro_to_toolbar(ut_tb_1)
            self.assertTrue(ii.ccc_called)
            self.assertIsNone(ii.pixmapText)



#***************************************************************************
#*   Copyright (c) 2018 Yorik van Havre <yorik@uncreated.net>              *
#*                                                                         *
#*   This program is free software; you can redistribute it and/or modify  *
#*   it under the terms of the GNU Lesser General Public License (LGPL)    *
#*   as published by the Free Software Foundation; either version 2 of     *
#*   the License, or (at your option) any later version.                   *
#*   for detail see the LICENCE text file.                                 *
#*                                                                         *
#*   This program is distributed in the hope that it will be useful,       *
#*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#*   GNU Library General Public License for more details.                  *
#*                                                                         *
#*   You should have received a copy of the GNU Library General Public     *
#*   License along with this program; if not, write to the Free Software   *
#*   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#*   USA                                                                   *
#*                                                                         *
#***************************************************************************


# This is the start page template. It builds an HTML global variable that
# contains the html code of the start page.
# Note: It is built only once per FreeCAD session for now...

import sys
import os
import tempfile
import time
import zipfile
import re
import FreeCAD
import FreeCADGui
import codecs
import hashlib
import urllib.parse
from . import TranslationTexts
from PySide import QtCore, QtGui

try:
    from addonmanager_macro import Macro as AM_Macro
    has_am_macro = True
except ImportError:
    has_am_macro = False


FreeCADGui.addLanguagePath(":/translations")
FreeCADGui.updateLocale()

iconprovider = QtGui.QFileIconProvider()
iconbank = {}       # store pre-existing icons so we don't overpollute temp dir
tempfolder = None   # store icons inside a subfolder in temp dir
defaulticon = None  # store a default icon for problematic file types


def getThumbnailDir():
    parent_dir = FreeCAD.getUserCachePath()
    return os.path.join(parent_dir, "FreeCADStartThumbnails")


def createThumbnailDir():
    path = getThumbnailDir()
    if not os.path.exists(path):
        os.mkdir(path)
    return path


def getSha1Hash(path, encode="utf-8"):
    sha_hash = hashlib.sha1()
    hashed = hashlib.sha1(path.encode(encode))
    hex_digest = hashed.hexdigest().encode(encode)
    sha_hash.update(hex_digest)
    return sha_hash.hexdigest()


def getUniquePNG(filename):
    parent_dir = getThumbnailDir()
    sha1 = getSha1Hash(filename) + ".png"
    return os.path.join(parent_dir, sha1)


def useCachedPNG(image, project):
    if not os.path.exists(image):
        return False
    if not os.path.exists(project):
        return False

    stamp = os.path.getmtime
    return stamp(image) > stamp(project)


def gethexcolor(color):

    "returns a color hex value #000000"

    r = str(hex(int(((color>>24)&0xFF))))[2:].zfill(2)
    g = str(hex(int(((color>>16)&0xFF))))[2:].zfill(2)
    b = str(hex(int(((color>>8)&0xFF))))[2:].zfill(2)
    return "#"+r+g+b



def isOpenableByFreeCAD(filename):

    "check if FreeCAD can handle this file type"

    if os.path.isdir(filename):
        return False
    if os.path.basename(filename)[0] == ".":
        return False
    extensions = [key.lower() for key in FreeCAD.getImportType().keys()]
    ext = os.path.splitext(filename)[1].lower()
    if ext:
        if ext[0] == ".":
            ext = ext[1:]
    if ext in extensions:
        return True
    return False



def getInfo(filename):

    "returns available file information"

    global iconbank,tempfolder

    tformat = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetString("TimeFormat","%m/%d/%Y %H:%M:%S")

    def getLocalTime(timestamp):
        "returns a local time from a timestamp"
        return time.strftime(tformat,time.localtime(timestamp))

    def getSize(size):
        "returns a human-readable size"
        if size > 1024*1024:
            hsize = str(int(size/(1024*1024))) + "Mb"
        elif size > 1024:
            hsize = str(int(size/1024)) + "Kb"
        else:
            hsize = str(int(size)) + "b"
        return hsize

    def getFreeDesktopThumbnail(filename):
        "if we have gnome libs available, try to find a system-generated thumbnail"
        path = os.path.abspath(filename)
        thumb = None
        try:
            import gnome.ui
            import gnomevfs
        except Exception:
            # alternative method
            import hashlib
            fhash = hashlib.md5(bytes(urllib.parse.quote("file://"+path,safe=":/"),"ascii")).hexdigest()
            thumb = os.path.join(os.path.expanduser("~"),".thumbnails","normal",fhash+".png")
        else:
            uri = gnomevfs.get_uri_from_local_path(path)
            thumb = gnome.ui.thumbnail_path_for_uri(uri, "normal")
        if thumb and os.path.exists(thumb):
            return thumb
        return None


    if os.path.exists(filename):

        if os.path.isdir(filename):
            return None

        # get normal file info
        s = os.stat(filename)
        size = getSize(s.st_size)
        ctime = getLocalTime(s.st_ctime)
        mtime = getLocalTime(s.st_mtime)
        author = ""
        company = TranslationTexts.T_UNKNOWN
        lic = TranslationTexts.T_UNKNOWN
        image = None
        descr = ""

        # get additional info from fcstd files
        if filename.lower().endswith(".fcstd"):
            try:
                zfile=zipfile.ZipFile(filename)
            except Exception:
                print("Cannot read file: ",filename)
                return None
            files=zfile.namelist()
            # check for meta-file if it's really a FreeCAD document
            if files[0] == "Document.xml":
                try:
                    doc = zfile.read(files[0]).decode('utf-8')
                except OSError as e:
                    print ("Fail to load corrupted FCStd file: '{0}' with this error: {1}".format(filename, str(e)))
                    return None
                doc = doc.replace("\n"," ")
                r = re.findall("Property name=\"CreatedBy.*?String value=\"(.*?)\"/>",doc)
                if r:
                    author = r[0]
                    # remove email if present in author field
                    if "&lt;" in author:
                        author = author.split("&lt;")[0].strip()
                r = re.findall("Property name=\"Company.*?String value=\"(.*?)\"/>",doc)
                if r:
                    company = r[0]
                r = re.findall("Property name=\"License.*?String value=\"(.*?)\"/>",doc)
                if r:
                    lic = r[0]
                r = re.findall("Property name=\"Comment.*?String value=\"(.*?)\"/>",doc)
                if r:
                    descr = r[0]
                if "thumbnails/Thumbnail.png" in files:
                    image_png = getUniquePNG(filename)
                    if filename in iconbank:
                        image = iconbank[filename]
                    elif useCachedPNG(image_png, filename):
                        image = image_png
                        iconbank[filename] = image
                    else:
                        imagedata=zfile.read("thumbnails/Thumbnail.png")
                        image = image_png
                        thumb = open(image,"wb")
                        thumb.write(imagedata)
                        thumb.close()
                        iconbank[filename] = image

        elif filename.lower().endswith(".fcmacro"):
            # For FreeCAD macros, use the Macro Editor icon (but we have to have it in a file for
            # the web view to load it)
            image = os.path.join(tempfolder,"fcmacro_icon.svg")
            if not os.path.exists(image):
                f = QtCore.QFile(":/icons/MacroEditor.svg")
                f.copy(image)
            iconbank[filename] = image

            if has_am_macro:
                macro = AM_Macro(os.path.basename(filename))
                macro.fill_details_from_file(filename)
                author = macro.author

        elif QtGui.QImageReader.imageFormat(filename):
            # use image itself as icon if it's an image file
            image = filename
            iconbank[filename] = image

        else:
            # use freedesktop thumbnail if available
            fdthumb = getFreeDesktopThumbnail(filename)
            if fdthumb:
                image = fdthumb
                iconbank[filename] = fdthumb

        # retrieve default mime icon if needed
        if not image:
            i = QtCore.QFileInfo(filename)
            t = iconprovider.type(i)
            filename_png = getUniquePNG(filename)
            if not t:
                t = "Unknown"
            if t in iconbank:
                image = iconbank[t]
            elif os.path.exists(filename_png):
                image = filename_png
                iconbank[t] = image
            else:
                icon = iconprovider.icon(i)
                if icon.availableSizes():
                    preferred = icon.actualSize(QtCore.QSize(128,128))
                    px = icon.pixmap(preferred)
                    image = filename_png
                    px.save(image)
                else:
                    image = getDefaultIcon()
                iconbank[t] = image
        return [image,size,author,ctime,mtime,descr,company,lic]

    return None



def getDefaultIcon():

    "retrieves or creates a default file icon"

    global defaulticon

    if not defaulticon:
        i = QtCore.QFileInfo(__file__) # MUST provide an existing file in qt5
        icon = iconprovider.icon(i)
        preferred = icon.actualSize(QtCore.QSize(128,128))
        px = icon.pixmap(preferred)
        image = getUniquePNG("default_icon")
        px.save(image)
        defaulticon = image

    return defaulticon



def buildCard(filename,method,arg=None):

    """builds an html <li> element representing a file.
    method is a script + a keyword, for ex. url.py?key="""

    result = ""
    if os.path.exists(filename) and isOpenableByFreeCAD(filename):
        basename = os.path.basename(filename)
        if not arg:
            arg = basename
        finfo = getInfo(filename)
        if finfo:
            image = finfo[0]
            size = finfo[1]
            author = finfo[2]
            infostring = TranslationTexts.T_CREATIONDATE+": "+finfo[3] + "\n"
            infostring += TranslationTexts.T_LASTMODIFIED+": "+finfo[4]
            if finfo[5]:
                infostring += "\n\n" + finfo[5]
            if size:
                result += '<li class="icon">'
                result += '<a href="'+method+urllib.parse.quote(arg)+'" title="'+infostring+'">'
                result += '<img src="file:///'+image.replace('\\','/')+'" alt="'+basename+'">'
                result += '<div class="caption">'
                result += '<h4>'+basename+'</h4>'
                result += '<p>'+author+'</p>'
                result += '<p>'+size+'</p>'
                result += '</div>'
                result += '</a>'
                result += '</li>'
    return result



def handle():

    "builds the HTML code of the start page"

    global iconbank,tempfolder

    # reuse stuff from previous runs to reduce temp dir clutter

    import Start
    if hasattr(Start,"iconbank"):
        iconbank = Start.iconbank
    if hasattr(Start,"tempfolder"):
        tempfolder = Start.tempfolder
    else:
        tempfolder = createThumbnailDir()

    # build the html page skeleton

    resources_dir = os.path.join(FreeCAD.getResourceDir(), "Mod", "Start", "StartPage")
    p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start")
    template = p.GetString("Template","")
    if template:
        html_filename = template
    else:
        html_filename = os.path.join(resources_dir, "StartPage.html")
    js_filename = os.path.join(resources_dir, "StartPage.js")
    css_filename = p.GetString("CSSFile",os.path.join(resources_dir, "StartPage.css"))
    with open(html_filename, 'r') as f:
        HTML = f.read()
    with open(js_filename, 'r') as f:
        JS = f.read()
    with open(css_filename, 'r') as f:
        CSS = f.read()
    HTML = HTML.replace("JS",JS)
    HTML = HTML.replace("CSS",CSS)

    # set the language

    HTML = HTML.replace("BCP47_LANGUAGE",QtCore.QLocale().bcp47Name())

    # get the stylesheet if we are using one

    if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetBool("UseStyleSheet",False):
        qssfile = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/MainWindow").GetString("StyleSheet","")
        if qssfile:
            # Search for stylesheet in user, system and resources locations
            user = os.path.join(FreeCAD.getUserAppDataDir(), "Gui", "Stylesheets")
            system = os.path.join(FreeCAD.getResourceDir(), "Gui", "Stylesheets")
            resources = ":/stylesheets"

            res = False
            if QtCore.QFile.exists(os.path.join(user, qssfile)):
                path = os.path.join(user, qssfile)
            elif QtCore.QFile.exists(os.path.join(system, qssfile)):
                path = os.path.join(system, qssfile)
            elif QtCore.QFile.exists(os.path.join(resources, qssfile)):
                res = True
                path = os.path.join(resources, qssfile)
            else:
                path = None

            if path:
                if res:
                    f = QtCore.QFile(path)
                    if f.open(QtCore.QIODevice.ReadOnly | QtCore.QFile.Text):
                        ALTCSS = QtCore.QTextStream(f).readAll()
                        HTML = HTML.replace("<!--QSS-->","<style type=\"text/css\">"+ALTCSS+"</style>")
                else:
                    with codecs.open(path, encoding='utf-8') as f:
                        ALTCSS = f.read()
                        HTML = HTML.replace("<!--QSS-->","<style type=\"text/css\">"+ALTCSS+"</style>")

    # turn tips off if needed

    if not FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetBool("ShowTips",True):
        HTML = HTML.replace("display: block; /* footnote tips display */","display: none; /* footnote tips display */")

    # get FreeCAD version

    v = FreeCAD.Version()
    VERSIONSTRING = TranslationTexts.T_VERSION + " " + v[0] + "." + v[1] + "." + v[2] +" " + TranslationTexts.T_BUILD + " " + v[3]
    HTML = HTML.replace("VERSIONSTRING",VERSIONSTRING)

    # translate texts

    texts = [t for t in dir(TranslationTexts) if t.startswith("T_")]
    for text in texts:
        HTML = HTML.replace(text,getattr(TranslationTexts,text))

    # build a "create new" icon with the FreeCAD background color gradient

    if not "createimg" in iconbank:
        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View")
        c1 = gethexcolor(p.GetUnsigned("BackgroundColor2"))
        c2 = gethexcolor(p.GetUnsigned("BackgroundColor3"))
        gradient = QtGui.QLinearGradient(0, 0, 0, 128)
        gradient.setColorAt(0.0, QtGui.QColor(c1))
        gradient.setColorAt(1.0, QtGui.QColor(c2))
        i = QtGui.QImage(128,128,QtGui.QImage.Format_RGB16)
        pa = QtGui.QPainter(i)
        pa.fillRect(i.rect(),gradient)
        pa.end()
        createimg = getUniquePNG("createimg")
        i.save(createimg)
        iconbank["createimg"] = createimg

    # build SECTION_RECENTFILES

    rf = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/RecentFiles")
    rfcount = rf.GetInt("RecentFiles",0)
    SECTION_RECENTFILES = "<h2>"+TranslationTexts.T_RECENTFILES+"</h2>"
    SECTION_RECENTFILES += "<ul>"
    SECTION_RECENTFILES += '<li class="icon">'
    SECTION_RECENTFILES += '<a href="LoadNew.py" title="'+TranslationTexts.T_CREATENEW+'">'
    if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetBool("NewFileGradient",False):
        SECTION_RECENTFILES += '<img src="file:///'+iconbank["createimg"].replace('\\','/')+'" alt="'+TranslationTexts.T_CREATENEW+'">'
    else:
        SECTION_RECENTFILES += '<img src="file:///'+os.path.join(resources_dir, "images/new_file_thumbnail.svg").replace('\\','/')+'" alt="'+TranslationTexts.T_CREATENEW+'">'
    SECTION_RECENTFILES += '<div class="caption">'
    SECTION_RECENTFILES += '<h4>'+TranslationTexts.T_CREATENEW+'</h4>'
    SECTION_RECENTFILES += '</div>'
    SECTION_RECENTFILES += '</a>'
    SECTION_RECENTFILES += '</li>'
    for i in range(rfcount):
        filename = rf.GetString("MRU%d" % (i))
        SECTION_RECENTFILES += buildCard(filename,method="LoadMRU.py?MRU=",arg=str(i))
    SECTION_RECENTFILES += '</ul>'
    HTML = HTML.replace("SECTION_RECENTFILES",SECTION_RECENTFILES)

    # build SECTION_EXAMPLES

    SECTION_EXAMPLES = ""
    if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetBool("ShowExamples",True):
        SECTION_EXAMPLES = "<h2>"+TranslationTexts.T_EXAMPLES+"</h2>"
        SECTION_EXAMPLES += "<ul>"
        examples_path = FreeCAD.getResourceDir()+"examples"
        if os.path.exists(examples_path):
            examples = os.listdir(examples_path)
            for basename in examples:
                filename = FreeCAD.getResourceDir()+"examples"+os.sep+basename
                SECTION_EXAMPLES += buildCard(filename,method="LoadExample.py?filename=")
        SECTION_EXAMPLES += "</ul>"
    HTML = HTML.replace("SECTION_EXAMPLES",SECTION_EXAMPLES)

    # build SECTION_CUSTOM

    SECTION_CUSTOM = ""
    cfolders = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetString("ShowCustomFolder","")
    if cfolders:
        dn = 0
        for cfolder in cfolders.split(";;"): # allow several paths separated by ;;
            if not os.path.isdir(cfolder):
                cfolder = os.path.dirname(cfolder)
            if not os.path.exists(cfolder):
                FreeCAD.Console.PrintWarning("Custom folder not found: %s" % cfolder)
            else:
                SECTION_CUSTOM += "<h2>"+os.path.basename(os.path.normpath(cfolder))+"</h2>"
                SECTION_CUSTOM += "<ul>"
                for basename in os.listdir(cfolder):
                    filename = os.path.join(cfolder,basename)
                    SECTION_CUSTOM += buildCard(filename,method="LoadCustom.py?filename="+str(dn)+"_")
                SECTION_CUSTOM += "</ul>"
                # hide the custom section tooltip if custom section is set (users know about it if they enabled it)
                HTML = HTML.replace("id=\"customtip\" class","id=\"customtip\" style=\"display:none;\" class")
                dn += 1
    HTML = HTML.replace("SECTION_CUSTOM",SECTION_CUSTOM)

    # build IMAGE_SRC paths

    HTML = HTML.replace("IMAGE_SRC_USERHUB",'file:///'+os.path.join(resources_dir, 'images/userhub.png').replace('\\','/'))
    HTML = HTML.replace("IMAGE_SRC_POWERHUB",'file:///'+os.path.join(resources_dir, 'images/poweruserhub.png').replace('\\','/'))
    HTML = HTML.replace("IMAGE_SRC_DEVHUB",'file:///'+os.path.join(resources_dir, 'images/developerhub.png').replace('\\','/'))
    HTML = HTML.replace("IMAGE_SRC_MANUAL",'file:///'+os.path.join(resources_dir, 'images/manual.png').replace('\\','/'))
    HTML = HTML.replace("IMAGE_SRC_SETTINGS",'file:///'+os.path.join(resources_dir, 'images/settings.png').replace('\\','/'))
    HTML = HTML.replace("IMAGE_SRC_INSTALLED",'file:///'+os.path.join(resources_dir, 'images/installed.png').replace('\\','/'))

    # build UL_WORKBENCHES

    wblist = []
    UL_WORKBENCHES = '<ul class="workbenches">'
    FreeCAD.getResourceDir()
    for wb in sorted(FreeCADGui.listWorkbenches().keys()):
        if wb.endswith("Workbench"):
            wn = wb[:-9]
        else:
            wn = wb
        # fixes for non-standard names
        if wn == "flamingoTools":
            wn = "flamingo"
        elif wn == "Geodat":
            wn = "geodata"
        elif wn == "a2p":
            wn = "A2plus"
        elif wn == "ArchTexture":
            wn = "ArchTextures"
        elif wn == "CadQuery":
            wn = "cadquery_module"
        elif wn == "DefeaturingWB":
            wn = "Defeaturing"
        elif wn == "ksuWB":
            wn = "kicadStepUp"
        elif wn == "ManipulatorWB":
            wn = "Manipulator"
        elif wn == "PartOMagic":
            wn = "Part-o-magic"
        elif wn == "SM":
            wn = "sheetmetal"
        elif wn == "gear":
            wn = "FCGear"
        elif wn == "frame_":
            wn = "frame"
        elif wn == "ReverseEngineering":
            wn = "Reverse_Engineering"
        elif wn == "None":
            continue
        wblist.append(wn.lower())
        if wb in iconbank:
            img = iconbank[wb]
        else:
            img = os.path.join(FreeCAD.getResourceDir(),"Mod",wn,"Resources","icons",wn+"Workbench.svg")
            if not os.path.exists(img):
                w = FreeCADGui.listWorkbenches()[wb]
                if hasattr(w,"Icon") and w.Icon:
                    xpm = w.Icon
                    if "XPM" in xpm:
                        xpm = xpm.replace("\n        ","\n") # some XPMs have some indent that QT doesn't like
                        r = [s[:-1].strip('"') for s in re.findall("(?s){(.*?)};",xpm)[0].split("\n")[1:]]
                        p = QtGui.QPixmap(r)
                        p = p.scaled(24,24)
                        img = getUniquePNG(wb)
                        p.save(img)
                    else:
                        img = xpm
                else:
                    img = os.path.join(resources_dir,"images/freecad.png")
            iconbank[wb] = img
        UL_WORKBENCHES += '<li>'
        UL_WORKBENCHES += '<img src="file:///'+img.replace('\\','/')+'" alt="'+wn+'">&nbsp;'
        UL_WORKBENCHES += '<a href="https://www.freecad.org/wiki/'+wn+'_Workbench">'+wn.replace("Reverse_Engineering","ReverseEng")+'</a>'
        UL_WORKBENCHES += '</li>'
    UL_WORKBENCHES += '</ul>'
    HTML = HTML.replace("UL_WORKBENCHES", UL_WORKBENCHES)

    # Detect additional addons that are not a workbench

    try:
        import dxfLibrary
    except Exception:
        pass
    else:
        wblist.append("dxf-library")
    try:
        import RebarTools
    except Exception:
        pass
    else:
        wblist.append("reinforcement")
    try:
        import CADExchangerIO
    except Exception:
        pass
    else:
        wblist.append("cadexchanger")
    HTML = HTML.replace("var wblist = [];","var wblist = " + str(wblist) + ";")

    # set and replace colors and font settings

    p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start")
    if p.GetString("BackgroundImage",""):
        BACKGROUND = gethexcolor(p.GetUnsigned("BackgroundColor1",1331197183))+" url("+p.GetString("BackgroundImage","")+")"
    else:
        BACKGROUND = gethexcolor(p.GetUnsigned("BackgroundColor1",1331197183))
        # linear gradient not supported by QT "linear-gradient("+gethexcolor(p.GetUnsigned("BackgroundColor1",1331197183))+","+gethexcolor(p.GetUnsigned("BackgroundColor2",2141107711))+")"
    LINKCOLOR = gethexcolor(p.GetUnsigned("LinkColor",65535))
    BASECOLOR = gethexcolor(p.GetUnsigned("PageColor",4294967295))
    BOXCOLOR  = gethexcolor(p.GetUnsigned("BoxColor",3722305023))
    TEXTCOLOR = gethexcolor(p.GetUnsigned("PageTextColor",255))
    BGTCOLOR = gethexcolor(p.GetUnsigned("BackgroundTextColor",4294703103))
    OVERFLOW = "" if p.GetBool("ShowScrollBars",True) else "body::-webkit-scrollbar {display: none;}"
    SHADOW = "#888888"
    if QtGui.QColor(BASECOLOR).valueF() < 0.5: # dark page - we need to make darker shadows
        SHADOW = "#000000"
    FONTFAMILY = p.GetString("FontFamily","Arial,Helvetica,sans")
    if not FONTFAMILY:
        FONTFAMILY = "Arial,Helvetica,sans"
    FONTSIZE = p.GetInt("FontSize",13)
    HTML = HTML.replace("BASECOLOR",BASECOLOR)
    HTML = HTML.replace("BOXCOLOR",BOXCOLOR)
    HTML = HTML.replace("LINKCOLOR",LINKCOLOR)
    HTML = HTML.replace("TEXTCOLOR",TEXTCOLOR)
    HTML = HTML.replace("BGTCOLOR",BGTCOLOR)
    HTML = HTML.replace("BACKGROUND",BACKGROUND)
    HTML = HTML.replace("SHADOW",SHADOW)
    HTML = HTML.replace("FONTFAMILY",FONTFAMILY)
    HTML = HTML.replace("FONTSIZE",str(FONTSIZE)+"px")
    HTML = HTML.replace("OVERFLOW",OVERFLOW)

    # enable web access if permitted

    if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetBool("AllowDownload",False):
        HTML = HTML.replace("var allowDownloads = 0;","var allowDownloads = 1;")

    # enable or disable forum

    if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetBool("ShowForum",False):
        HTML = HTML.replace("var showForum = 0;","var showForum = 1;")
        HTML = HTML.replace("display: none; /* forum display */","display: block; /* forum display */")

    # enable or disable notepad

    if FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start").GetBool("ShowNotes",False):
        HTML = HTML.replace("display: none; /* notes display */","display: block; /* notes display */")
        HTML = HTML.replace("width: 100%; /* thumbs display */","width: 70%; /* thumbs display */")

    # store variables for further use

    Start.iconbank = iconbank
    Start.tempfolder = tempfolder

    return HTML



def exportTestFile():

    "Allow to check if everything is Ok"

    with codecs.open(
        os.path.expanduser("~") + os.sep + "freecad-startpage.html",
        encoding="utf-8",
        mode="w",
    ) as f:
        f.write(handle())
        f.close()



def postStart():

    "executes needed operations after loading a file"

    param = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/Start")

    # switch workbench
    wb = param.GetString("AutoloadModule","")
    if "$LastModule" == wb:
        wb = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/General").GetString("LastModule","")
    if wb:
        # don't switch workbenches if we are not in Start anymore
        if FreeCADGui.activeWorkbench() and (FreeCADGui.activeWorkbench().name() == "StartWorkbench"):
            FreeCADGui.activateWorkbench(wb)

    # close start tab
    cl = param.GetBool("closeStart",False)
    if cl:
        title = QtGui.QApplication.translate("Workbench","Start page")
        mw = FreeCADGui.getMainWindow()
        if mw:
            mdi = mw.findChild(QtGui.QMdiArea)
            if mdi:
                for mdichild in mdi.children():
                    for subw in mdichild.findChildren(QtGui.QMdiSubWindow):
                        if subw.windowTitle() == title:
                            subw.close()



def checkPostOpenStartPage():

    "on Start WB startup, check if we are loading a file and therefore need to close the StartPage"

    import Start
    if FreeCAD.ParamGet('User parameter:BaseApp/Preferences/Mod/Start').GetBool('DoNotShowOnOpen',False) and (not hasattr(Start,'CanOpenStartPage')):
        if len(sys.argv) > 1:
            postStart()
    Start.CanOpenStartPage = True
